/*
 * Copyright (c) 2002-2006, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* readfile.cpp */

/* Contains routines needed to read data from files */

#include "readfile.h"
#include "winExtra.h"

void Parameters::Read( FILE* input ) {
  /* This procedure reads in the parameters from the file */

  char* param_name;
  char* temp_char;

  /* Index variable(s) */
  long i;

  line_number=1;
  EOF_TRUE=0;
  
  do
    {
      param_name=read_word( input );
      if( !EOF_TRUE )
    	{
	  if( strcmp( param_name, "num_variables" )==0 )
	    { /* Reading in the number of states */
	      num_variables=read_long( input );
	      if( num_variables<1 )
		{
		  fprintf( stderr, "Error on line %ld: number of hidden variables must be at least 1. Setting it to 1.\n",
			   line_number );
		  num_variables=1;
		}
	      else if( num_variables>MAXNUMVARIABLES )
		{
		  fprintf( stderr, "Error on line %ld: number of hidden variables has a ceiling of %ld. Setting it to the ceiling.\n",
			   line_number, (long)MAXNUMVARIABLES );
		  num_variables=MAXNUMVARIABLES;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of latent variables: %ld\n", num_variables );
#endif
	    } /* Reading in the number of states */
	  if( strcmp( param_name, "num_states" )==0 )
	    { /* Reading in the number of states */
	      num_states=read_long( input );
	      if( num_states<1 )
		{
		  fprintf( stderr, "Error on line %ld: number of hidden states must be at least 1. Aborting.\n",
			   line_number );
		  exit( -1 );
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of states: %ld\n", num_states );
#endif
	    } /* Reading in the number of states */
	  else if( strcmp(param_name, "model_type" )==0 )
	    { /* Reading in the type of the model */
	      temp_char=read_word( input );
	      if( strcmp( temp_char, "hmm" )==0 )
		model_type=MODEL_TYPE_HMM;
	      else if( strcmp( temp_char, "nhmm" )==0 )
		model_type=MODEL_TYPE_NHMM;
	      else if( strcmp( temp_char, "mixture" )==0 )
		model_type=MODEL_TYPE_MIX;
	      else if( strcmp( temp_char, "nmixture" )==0 )
		model_type=MODEL_TYPE_NMIX;
	      else
		model_type=MODEL_TYPE_HMM;
	      
	      delete temp_char;
#ifdef INPUT_VERBOSE
	      printf( "Type of the model: " );
	      switch( model_type ) {
	      case MODEL_TYPE_HMM:
		printf( "HMM\n" );
		break;
	      case MODEL_TYPE_NHMM:
		printf( "NHMM\n" );
		break;
	      case MODEL_TYPE_MIX:
		printf( "mixture\n" );
		break;
	      case MODEL_TYPE_NMIX:
		printf( "non-homogeneous mixture\n" );
		break;
	      default:
		printf( "unknown\n" );
	      }
#endif
	    } /* Reading in the type of the model */
	  else if( strcmp( param_name, "action" )==0 )
	    { /* Action */
	      temp_char=read_word( input );
	      if( strcmp( temp_char, "learn" )==0 )
		/* Estimate parameters */
		action=ACTION_CODE_LEARN;
	      else if( strcmp( temp_char, "viterbi" )==0 )
		/* Find most likely sequence of states */
		action=ACTION_CODE_VITERBI;
	      else if( strcmp( temp_char, "ll" )==0 )
		/* Calculate log-likelihood of the model given the data */
		action=ACTION_CODE_LL;
	      else if( strcmp( temp_char, "ll-train" )==0 )
		/* Calculating log-likelihood of the training set given the model */
		action=ACTION_CODE_LLTRAIN;
	      else if( strcmp( temp_char, "simulation" )==0 )
		/* Simulate the data from the model (and, possibly, inputs) */
		action=ACTION_CODE_SIM;
	      else if( strcmp( temp_char, "analysis" )==0 )
		/* Analyze the data */
		action=ACTION_CODE_ANALYZE;
	      else if( strcmp( temp_char, "filling" )==0 )
		/* Hole poking evaluation */
		action=ACTION_CODE_FILLING;
	      else if( strcmp( temp_char, "sim-filling" )==0 )
		/* Simulating missing entries */
		action=ACTION_CODE_SIM_FILL;
	      else if( strcmp( temp_char, "prediction" )==0 )
		/* Computing predictive ability */
		action=ACTION_CODE_PREDICT;
	      else if( strcmp( temp_char, "init" )==0 )
		/* Creating initial sets of parameters */
		action=ACTION_CODE_INIT;
	      else if( strcmp( temp_char, "KL" )==0 )
		/* Computing Kullback-Leibler divergence */
		action=ACTION_CODE_KL;
	      else if( strcmp( temp_char, "debug" )==0 )
		/* Testing recently implemented functionality */
		action=ACTION_CODE_DEBUG;
	      else
		/* None of the above */
 		action=ACTION_CODE_UNKNOWN;

	      delete temp_char;
#ifdef INPUT_VERBOSE
	      printf( "Action: " );
	      switch( action ) {
	      case ACTION_CODE_LEARN:
		printf( "learning\n" );
		break;
	      case ACTION_CODE_VITERBI:
		printf( "viterbi\n" );
		break;
	      case ACTION_CODE_LL:
		printf( "log-likelihood\n" );
		break;
	      case ACTION_CODE_LLTRAIN:
		printf( "log-likelihood of the training set\n" );
		break;
	      case ACTION_CODE_SIM:
		printf( "simulation\n" );
		break;
	      case ACTION_CODE_ANALYZE:
		printf( "data analysis\n" );
		break;
	      case ACTION_CODE_FILLING:
		printf( "hole filling evaluation\n" );
		break;
	      case ACTION_CODE_SIM_FILL:
		printf( "simulation of data for missing data\n" );
		break;
	      case ACTION_CODE_PREDICT:
		printf( "prediction\n" );
		break;
	      case ACTION_CODE_INIT:
		printf( "parameter initialization\n" );
		break;
	      case ACTION_CODE_KL:
		printf( "Kullback-Leibler divergence\n" );
		break;
	      case ACTION_CODE_DEBUG:
		printf( "debugging\n" );
		break;
	      default:
		printf( "unknown\n" );
	      }
#endif
	    } /* Action */
	  else if( strcmp( param_name, "emission" )==0 )
	    { /* Reading in the emission parameters */
#ifdef INPUT_VERBOSE
	      printf( "Reading in the emission distributions' specifications for each of %ld variables.\n", num_variables );
#endif
	      for( i=0; i<num_variables; i++ )
		{
#ifdef INPUT_VERBOSE
		  printf( "Distribution for variable %ld:\n", i+1 );
#endif
		  if( emission[i] )
		    delete emission[i];
		  emission[i]=ReadDistribution( input );
		}
	    } /* Reading in the emission parameters */
	  else if( strcmp( param_name, "transition" )==0 )
	    { /* Reading in the transition distribution */
	      if( transition )
		delete transition;
	      transition=ReadDistribution( input );
	    } /* Reading in the transition distribution */
	  else if( strcmp( param_name, "data" )==0 )
	    { /* Data filename */
	      if( output_data_filename )
		delete output_data_filename;
	      output_data_filename=read_word( input );
#if defined  WIN32 || CYGWIN
	      /* remove " from file name */
	      getWindowFileName(output_data_filename);
#endif

#ifdef INPUT_VERBOSE
	      printf( "Data file name: %s\n", output_data_filename );
#endif
	    } /* Data filename */
	  else if( strcmp( param_name, "num_discrete_data_components" )==0 )
	    { /* Number of finite-valued components in the output data */
	      num_ddata_components=read_long( input );
	      if( num_ddata_components<0 )
		{
		  fprintf( stderr, "Warning on line %ld: number of output data finite-valued components must be at non-negative.  Setting to 0.\n",
			   line_number );
		  num_ddata_components=0;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of finite-valued vector data components: %ld\n", num_ddata_components );
#endif
	    } /* Number of finite-valued components in the output data */
	  else if( strcmp( param_name, "num_real_data_components" )==0 )
	    { /* Number of real-valued components in the output data */
	      num_rdata_components=read_long( input );
	      if( num_rdata_components<0 )
		{
		  fprintf( stderr, "Warning on line %ld: number of output data real-valued components must be at non-negative.  Setting to 0.\n",
			   line_number );
		  num_rdata_components=0;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of real-valued vector data components: %ld\n", num_rdata_components );
#endif
	    } /* Number of real-valued components in the output data */
	  else if( strcmp( param_name, "num_data_sequences" )==0 )
	    { /* Number of data sequences */
	      num_data_seqs=read_long( input );
	      if( num_data_seqs<1 )
		{
		  fprintf( stderr, "Error on line %ld: number of data sequences must be at least 1.  Aborting.\n",
			   line_number );
		  exit( -1 );
		}
	      if( length_data_seq )
		{
		  delete length_data_seq;
		  length_data_seq=NULL;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of data sequences: %ld\n", num_data_seqs );
#endif
	    } /* Number of data sequences */
	  else if( strcmp( param_name, "data_sequence_length" )==0 )
	    { /* Lengths of each data sequence */
	      /* !!! Assumes that lengths of all sequences are the same !!! */
	      
	      /* Reading in the number of entries in each sequence */
	      length_data_seqs=read_long( input );
	      if( length_data_seqs<1 )
		{
		  fprintf( stderr, "Error on line %ld: lengths of data sequences must be at least 1.  Aborting.\n",
			   line_number );
		  exit( -1 );		  
		}
	      if( length_data_seq )
		{
		  delete length_data_seq;
		  length_data_seq=NULL;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of data points per sequence: %ld\n", length_data_seqs );
#endif
	    } /* Lengths of each data sequence */
	  else if( strcmp( param_name, "data_sequence_length_distinct" )==0 )
	    { /* Possibly different length for each data sequence */
	      if( length_data_seq )
		delete length_data_seq;
	      length_data_seq=new long[num_data_seqs];
	      for( i=0; i<num_data_seqs; i++ )
		{
		  length_data_seq[i]=read_long( input );
		  if( length_data_seq[i]<1 )
		    {
		      fprintf( stderr, "Error on line %ld: length of data sequences must be at least 1. Aborting.\n",
			       line_number );
		      exit( -1 );
		    }
		}
#ifdef INPUT_VERBOSE
	      for( i=0; i<num_data_seqs; i++ )
		printf( "Number of data point for data sequence %ld: %ld\n", i+1, length_data_seq[i] );
#endif
	    } /* Possibly different length for each data sequence */
	  else if( strcmp( param_name, "input_dimensionality" )==0 )
	    { /* Reading in input distribution */
#ifdef INPUT_VERBOSE
	      printf( "Reading in the number of input components:\n" );
#endif
	      input_dim=read_long( input );
	      if( input_dim<0 )
		{
		  fprintf( stderr, "Error on line %ld: number of input components must be at least 0.  Resetting to 0.\n",
			   line_number );
		  input_dim=0;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of input components: %ld\n", input_dim );
#endif
	    } /* Reading in input distribution */
	  else if( strcmp( param_name, "input_filename" )==0 )
	    { /* Input data filename */
	      if( input_filename )
		delete input_filename;
	      input_filename=read_word( input );
#if defined WIN32 || CYGWIN
	      /* remove " from file name */
	      getWindowFileName(input_filename);
#endif

#ifdef INPUT_VERBOSE
	      printf( "Input file name : %s\n", input_filename );
#endif
	    } /* Input data filename */
	  else if( strcmp( param_name, "num_discrete_input_components" )==0 )
	    { /* Number of finite-valued components in the input data */
	      num_dinput_components=read_long( input );
	      if( num_dinput_components<0 )
		{
		  fprintf( stderr, "Warning on line %ld: number of input data discrete-valued components must be at non-negative.  Setting to 0.\n",
			   line_number );
		  num_dinput_components=0;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of finite-valued input vector data components: %ld\n", num_dinput_components );
#endif
	    } /* Number of finite-valued components in the input data */
	  else if( strcmp( param_name, "num_real_input_components" )==0 )
	    { /* Number of real-valued components in the input data */
	      num_rinput_components=read_long( input );
	      if( num_rinput_components<0 )
		{
		  fprintf( stderr, "Warning on line %ld: number of input data real-valued components must be at non-negative.  Setting to 0.\n",
			   line_number );
		  num_rinput_components=0;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of real-valued vector input data components: %ld\n", num_rinput_components );
#endif
	    } /* Number of real-valued components in the input data */
	  else if( strcmp( param_name, "num_input_sequences" )==0 )
	    { /* Number of input sequences */
	      num_input_seqs=read_long( input );
	      if( num_input_seqs<1 )
		{
		  fprintf( stderr, "Error on line %ld: number of input sequences must be at least 1.  Aborting.\n",
			   line_number );
		  exit( -1 );
		}
	      if( length_input_seq )
		{
		  delete length_input_seq;
		  length_input_seq=NULL;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of input sequences: %ld\n", num_input_seqs );
#endif
	    } /* Number of input sequences */
	  else if( strcmp( param_name, "input_sequence_length" )==0 )
	    { /* Lengths of each input sequence */
	      /* !!! Assumes that lengths of all sequences are the same !!! */
	      
	      /* Reading in the number of entries in each sequence */
	      length_input_seqs=read_long( input );
	      if( length_input_seqs<1 )
		{
		  fprintf( stderr, "Error on line %ld: lengths of input sequences must be at least 1.  Aborting.\n",
			   line_number );
		  exit( -1 );		  
		}
	      if( length_input_seq )
		{
		  delete length_input_seq;
		  length_input_seq=NULL;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of data points per input sequence: %ld\n", length_input_seqs );
#endif
	    } /* Lengths of each input sequence */
	  else if( strcmp( param_name, "input_sequence_length_distinct" )==0 )
	    { /* Possibly different length for each input sequence */
	      if( length_input_seq )
		delete length_input_seq;
	      length_input_seq=new long[num_input_seqs];
	      for( i=0; i<num_input_seqs; i++ )
		{
		  length_input_seq[i]=read_long( input );
		  if( length_input_seq[i]<1 )
		    {
		      fprintf( stderr, "Error on line %ld: length of input sequences must be at least 1. Aborting.\n",
			       line_number );
		      exit( -1 );
		    }
		}
#ifdef INPUT_VERBOSE
	      for( i=0; i<num_input_seqs; i++ )
		printf( "Number of data point for input sequence %ld: %ld\n", i+1, length_input_seq[i] );
#endif
	    } /* Possibly different length for each input sequence */
	  else if( strcmp( param_name, "extra_filename" )==0 )
	    { /* Extra data filename */
	      if( extra_data_filename )
		delete extra_data_filename;
	      extra_data_filename=read_word( input );

#if defined WIN32 || CYGWIN
	      /* remove " from file name */
	      getWindowFileName(extra_data_filename);
#endif

#ifdef INPUT_VERBOSE
	      printf( "Extra data file name : %s\n", extra_data_filename );
#endif
	    } /* Extra data filename */
	  else if( strcmp( param_name, "num_discrete_extra_components" )==0 )
	    { /* Number of finite-valued components in the extra data */
	      num_dextra_components=read_long( input );
	      if( num_dextra_components<0 )
		{
		  fprintf( stderr, "Warning on line %ld: number of extra data discrete-valued components must be at non-negative.  Setting to 0.\n",
			   line_number );
		  num_dextra_components=0;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of finite-valued extra vector data components: %ld\n", num_dextra_components );
#endif
	    } /* Number of finite-valued components in the extra data */
	  else if( strcmp( param_name, "num_real_extra_components" )==0 )
	    { /* Number of real-valued components in the extra data */
	      num_rextra_components=read_long( input );
	      if( num_rextra_components<0 )
		{
		  fprintf( stderr, "Warning on line %ld: number of extra data real-valued components must be at non-negative.  Setting to 0.\n",
			   line_number );
		  num_rextra_components=0;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of real-valued vector extra data components: %ld\n", num_rextra_components );
#endif
	    } /* Number of real-valued components in the extra data */
	  else if( strcmp( param_name, "num_extra_sequences" )==0 )
	    { /* Number of input sequences */
	      num_extra_seqs=read_long( input );
	      if( num_extra_seqs<1 )
		{
		  fprintf( stderr, "Error on line %ld: number of extra sequences must be at least 1.  Aborting.\n",
			   line_number );
		  exit( -1 );
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of extra sequences: %ld\n", num_extra_seqs );
#endif
	    } /* Number of extra sequences */
	  else if( strcmp( param_name, "extra_sequence_length" )==0 )
	    { /* Lengths of each extra sequence */
	      /* !!! Assumes that lengths of all sequences are the same !!! */
	      
	      /* Reading in the number of entries in each sequence */
	      length_extra_seqs=read_long( input );
	      if( length_extra_seqs<1 )
		{
		  fprintf( stderr, "Error on line %ld: lengths of extra sequences must be at least 1.  Aborting.\n",
			   line_number );
		  exit( -1 );		  
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of data points per extra sequence: %ld\n", length_extra_seqs );
#endif
	    } /* Lengths of each extra sequence */
	  else if( strcmp( param_name, "xval_type" )==0 )
	    { /* Type of cross-validation */
	      temp_char=read_word( input );
	      if( strcmp( temp_char, "none" )==0 )
		/* No cross-validation */
		xval_type=XVAL_NONE;
	      else if( strcmp( temp_char, "leave_n_out" )==0 )
		/* Leave-n-out cross-validation */
		xval_type=XVAL_LEAVENOUT;
	      else
		/* Default */
 		xval_type=XVAL_NONE;

	      delete temp_char;
#ifdef INPUT_VERBOSE
	      printf( "Cross-validation type: " );
	      switch( xval_type ) {
	      case XVAL_NONE:
		printf( "none\n" );
		break;
	      case XVAL_LEAVENOUT:
		printf( "leave-n-out\n" );
		break;
	      default:
		printf( "unknown\n" );
	      }
#endif	      
	    } /* Type of cross-validation */
	  else if( strcmp( param_name, "examples_out" )==0 )
	    { /* Number of examples to leave out for cross-validation */
	      number_out=read_long( input );
	      if( number_out<1 )
		{
		  fprintf( stderr, "Number of examples to leave out is too small.  Setting it to 1.\n" );
		  number_out=1;
		}	      
#ifdef INPUT_VERBOSE
	      printf( "Number of examples to leave out: %ld\n", number_out );
#endif
	    } /* Number of examples to leave out for cross-validation */
	  else if( strcmp( param_name, "model_filename" )==0 )
	    { /* Model */

	      /* Reading in the filename of the file with the model */
	      if( model_filename )
		delete model_filename;
	      model_filename=read_word( input );

#if defined WIN32 || CYGWIN
	      /* remove " from file name */
	      getWindowFileName(model_filename);
#endif

#ifdef INPUT_VERBOSE
	      printf( "Model filename %s\n", model_filename );
#endif
	    } /* Model */
	  else if( strcmp( param_name, "num_models" )==0 )
	    { /* Number of models in the model file */
	      num_models=read_long( input );
	      if( num_models<1 )
		{
		  fprintf( stderr, "Warning on line %ld: number of models in the model file must be at least 1.  Setting to 1.\n",
			   line_number );
		  num_models=1;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of models in the model file: %ld (may be adjusted later)\n", num_models );
#endif
	    } /* Number of models in the model file */
	  else if( strcmp( param_name, "output" )==0 )
	    { /* Output file */
	      output_filename=read_word( input );
#if defined  WIN32 || CYGWIN
		/* remove " from file name */
		getWindowFileName(output_filename);
#endif

#ifdef INPUT_VERBOSE
	      printf( "Output filename: %s\n", output_filename );
#endif
	    } /* Output file */
	  else if( strcmp( param_name, "num_restarts" )==0 )
	    { /* Number of runs for EM */
	      num_restarts=read_long( input );
	      if( num_restarts<1 )
		{
		  fprintf( stderr, "Warning on line %ld: number of random restarts must be at least 1.  Setting to 1.\n",
			   line_number );
		  num_restarts=1;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of EM random restarts: %ld\n", num_restarts );
#endif
	    } /* Number of runs for EM */
	  else if( strcmp( param_name, "em_precision" )==0 ) 
	    { /* Precision of EM algorithm */
	      em_precision=read_double( input );
	      if( em_precision<0.0 )
		{
		  fprintf( stderr, "Warning on line %ld: EM precision must be positive.  Resetting to default.\n",
			   line_number );
		  em_precision=EM_EPSILON;
		}
#ifdef INPUT_VERBOSE
	      printf( "EM with precision %0.12le\n", em_precision );
#endif
	    } /* Precision of EM algorithm */
	  else if( strcmp( param_name, "em_verbose" )==0 )
	    { /* Flag whether to display verbose messages with EM */
	      em_verbose=1;
#ifdef INPUT_VERBOSE
	      printf( "Displaying intermediate messages with EM\n" );
#endif
	    } /* Flag whether to display verbose messages with EM */
	  else if( strcmp( param_name, "initialization" )==0 )
	    { /* Initialization type */
	      temp_char=read_word( input );
	      if( strcmp( temp_char, "random" )==0 )
		/* Random initial parameter assignment */
		initialization_type=INIT_RANDOM;
	      else if( strcmp( temp_char, "em" )==0 )
		/* EM on a simpler model */
		initialization_type=INIT_EM;
	      else
		/* Default */
 		initialization_type=INIT_RANDOM;

	      delete temp_char;
#ifdef INPUT_VERBOSE
	      printf( "Model parameter initialization type: " );
	      switch( initialization_type ) {
	      case INIT_RANDOM:
		printf( "random\n" );
		break;
	      case INIT_EM:
		printf( "EM on a simpler model\n" );
		break;
	      default:
		printf( "unknown\n" );
	      }
#endif	      	      
	    } /* Initialization type */
	  else if( strcmp( param_name, "num_simulations" )==0 )
	    { /* Number of simulations per run */
	      num_simulations=read_long( input );
	      if( num_simulations<1 )
		{
		  fprintf( stderr, "Warning on line %ld: number of simulated sequences must be at least 1.  Setting to 1.\n",
			   line_number );
		  num_simulations=1;
		}
#ifdef INPUT_VERBOSE
	      printf( "Number of simulated sequences: %ld\n", num_simulations );
#endif
	    } /* Number of simulations per run */
	  else if( strcmp( param_name, "state_filename" )==0 )
	    { /* State filename */
	      if( state_filename )
		delete state_filename;
	      state_filename=read_word( input );
#if defined WIN32 || CYGWIN
		/* remove " from file name */
		getWindowFileName(state_filename);
#endif

#ifdef INPUT_VERBOSE
	      printf( "Filename for the states %s\n", state_filename );
#endif
	    } /* State filename */
	  else if( strcmp( param_name, "analysis" )==0 )
	    { /* Type of analysis */
	      temp_char=read_word( input );
	      if( strcmp( temp_char, "mean" )==0 )
		/* Mean */
		analysis_type=ANALYSIS_MEAN;
	      else if( strcmp( temp_char, "correlation" )==0 )
		/* Correlation */
		analysis_type=ANALYSIS_CORR;
	      else if( strcmp( temp_char, "persistence" )==0 )
		/* Persistence */
		analysis_type=ANALYSIS_PERS;
	      else if( strcmp( temp_char, "dry" )==0 )
		/* Dry spell */
		analysis_type=ANALYSIS_DRY;
	      else if( strcmp( temp_char, "wet" )==0 )
		/* Wet spell */
		analysis_type=ANALYSIS_WET;
	      else if( strcmp( temp_char, "information" )==0 )
		/* Mutual information */
		analysis_type=ANALYSIS_MI;
	      else if( strcmp( temp_char, "logodds" )==0 )
		/* Log-odds ratio */
		analysis_type=ANALYSIS_LO;
	      else if( strcmp( temp_char, "comparison" )==0 )
		/* Comparison of true set to the filled-in set */
		analysis_type=ANALYSIS_COMP;
	      else
		/* Default */
 		analysis_type=ANALYSIS_MEAN;

	      delete temp_char;
#ifdef INPUT_VERBOSE
	      printf( "Analysis type: " );
	      switch( analysis_type ) {
	      case ANALYSIS_MEAN:
		printf( "mean\n" );
		break;
	      case ANALYSIS_CORR:
		printf( "correlation\n" );
		break;
	      case ANALYSIS_PERS:
		printf( "persistence\n" );
		break;
	      case ANALYSIS_DRY:
		printf( "dry spell length\n" );
		break;
	      case ANALYSIS_WET:
		printf( "wet spell length\n" );
		break;
	      case ANALYSIS_MI:
		printf( "mutual information\n" );
		break;
	      case ANALYSIS_LO:
		printf( "log-odds ratio\n" );
		break;
	      case ANALYSIS_COMP:
		printf( "comparison\n" );
		break;
	      default:
		printf( "unknown\n" );
	      }
#endif	      
	    } /* Type of analysis */
	  else if( strcmp( param_name, "filling" )==0 )
	    { /* Type of hole-filling analysis */
	      temp_char=read_word( input );
	      if( strcmp( temp_char, "log_p" )==0 )
		/* Log-probability */
		poking_type=POKING_TYPE_LOGP;
	      else if( strcmp( temp_char, "prediction" )==0 )
		/* Prediction */
		poking_type=POKING_TYPE_PRED;
	      else if( strcmp( temp_char, "missing-probabilities" )==0 )
		/* Marginal probabilities for missing values */
		poking_type=MISSING_IND_PROB;
	      else if( strcmp( temp_char, "missing-predictions" )==0 )
		/* Most likely predictions for missing probabilities */
		poking_type=MISSING_IND_PRED;
	      else if( strcmp( temp_char, "hidden-states" )==0 )
		/* Probabilities for hidden states given the data */
		poking_type=HIDDEN_STATE_PROB;
	      else if( strcmp( temp_char, "fill-in" )==0 )
		/* Filling with maximum probability values */
		poking_type=VITERBI_FILL_IN;
	      else
		/* Default */
 		poking_type=POKING_TYPE_LOGP;

	      delete temp_char;
#ifdef INPUT_VERBOSE
	      printf( "Hole-filling analysis type: " );
	      switch( poking_type ) {
	      case POKING_TYPE_LOGP:
		printf( "log-probability\n" );
		break;
	      case POKING_TYPE_PRED:
		printf( "prediction accuracy\n" );
		break;
	      case MISSING_IND_PROB:
		printf( "marginal probabilities for missng values\n" );
		break;		
	      case MISSING_IND_PRED:
		printf( "marginal best-probability predictions for missing values\n" );
		break;
	      case HIDDEN_STATE_PROB:
		printf( "probabilities for hidden states\n" );
		break;
	      case VITERBI_FILL_IN:
		printf( "filling in with maximum probability values\n" );
		break;
	      default:
		printf( "unknown\n" );
	      }
#endif	      
	    } /* Type of hole-filling analysis */
	  else if( strcmp( param_name, "robust_first_state" )==0 )
	    { /* Robust first state */
	      robust_first_state=1;
#ifdef INPUT_VERBOSE
	      printf( "Using robust first state probabilities estimation\n" );
#endif
	    } /* Robust first state */
	  else if( strcmp( param_name, "cg-epsilon" )==0 )
	    { /* Minimum conjugate gradient change */
	      cg_epsilon=read_double( input );
	      if( cg_epsilon<0.0 )
		{
		  fprintf( stderr, "Warning on line %ld: minimum epsilon for conjugate gradient is negative.  Resetting to default.\n",
			   line_number );
		  cg_epsilon=CONJ_GRAD_EPSILON;
		}
#ifdef INPUT_VERBOSE
	      printf( "Epsilon for conjugate gradient: %le\n", cg_epsilon );
#endif
	    } /* Minimum conjugate gradient change */
	  else if( strcmp( param_name, "maxent-epsilon" )==0 )
	    { /* Minimum maxent log-likelihood change */
	      maxent_epsilon=read_double( input );
	      if( maxent_epsilon<0.0 )
		{
		  fprintf( stderr, "Warning on line %ld: minimum epsilon for maxent log-likelihood is negative.  Resetting to default.\n",
			   line_number );
		  maxent_epsilon=MIN_ME_LL_CHANGE;
		}
#ifdef INPUT_VERBOSE
	      printf( "Epsilon for maxent log-likelihood: %le\n", cg_epsilon );
#endif
	    } /* Minimum maxent log-likelihood change */
	  else if( strcmp( param_name, "lookahead" )==0 )
	    { /* Look-ahead */
	      lookahead=read_long( input );
	      if( lookahead<1 )
		{
		  fprintf( stderr, "Warning on line %ld: lookahead is less than 1.  Resetting to default\n", line_number );
		  lookahead=1;
		}
	    } /* Look-ahead */
	  else if( strcmp( param_name, "short-display" )==0 )
	    /* !!! DEFAULT !!! */
	    { /* Not outputting/reading dimension indices for distributions */
	      short_dist_display=1;
#ifdef INPUT_VERBOSE
	      printf( "Not outputting dimension indices for distributions\n" );
#endif
	    } /* Not outputting/reading dimension indices for distributions */
	  else if( strcmp( param_name, "dim-index-display" )==0 )
	    { /* Outputting/reading dimension indices for distributions */
	      short_dist_display=0;
#ifdef INPUT_VERBOSE
	      printf( "Outputting dimension indices for distributions\n" );
#endif
	    } /* Outputting/reading dimension indices for distributions */
	  else if( strcmp( param_name, "bare-display" )==0 )
	    { /* Outputting parameters only, no comments */
	      bare_display=1;
#ifdef INPUT_VERBOSE
	      printf( "Outtping parameters only, no comments\n" );
#endif
	    } /* Outputting parameters only, no comments */
	  else
	    fprintf( stderr, "Unknown parameter name on line %ld: %s\nSkipping\n", line_number, param_name );  
	}

      delete param_name;
    }
  while( !EOF_TRUE );  

  return;
}

void DataPoint::ReadData( FILE* input ) {
  /* Index variable(s) */
  long i;
  
  for( i=0; i<ddim; i++ )
    ddata[i]=read_long( input );

  for( i=0; i<rdim; i++ )
    rdata[i]=read_double( input );

  return;
}


void Sequence::ReadData( FILE* input, DataPoint* datum ) {
  /* Index variable(s) */
  long i;

  for( i=0; i<seq_length; i++ )
    {
      /* Initializing and reading in the observation i */
      if( entry[i] )
	delete entry[i];
      entry[i]=new DataPoint( datum->ddim, datum->rdim );
      entry[i]->ReadData( input );
    }

  return;
}

void Data::ReadData( FILE* data_file, long start_position, long end_position, long seq_length, DataPoint* datum ) {
  /* This routine reads the data from a file into a structure */

  /* Index variable(s) */
  long i;
  
  line_number=1;
  EOF_TRUE=0;
  
  for( i=start_position; i<end_position; i++ )
    {
      /* Initializing and reading in the sequence i */
      if( sequence[i] )
	delete sequence[i];
      sequence[i]=new Sequence( seq_length );
      sequence[i]->ReadData( data_file, datum );
#ifdef INPUT_VERBOSE
      printf( "Read sequence #%d\n", i+1 );
#endif
    }
  
  return;
}

void Data::ReadDataDistinct( FILE* data_file, long start_position, long end_position, long* seq_length, DataPoint* datum ) {
  /* This routine reads the data from a file into a structure */

  /* Index variable(s) */
  long i;
  
  line_number=1;
  EOF_TRUE=0;
  
  for( i=start_position; i<end_position; i++ )
    {
      /* Initializing and reading in the sequence i */
      if( sequence[i] )
	delete sequence[i];
      sequence[i]=new Sequence( seq_length[i] );
      sequence[i]->ReadData( data_file, datum );
#ifdef INPUT_VERBOSE
      printf( "Read sequence #%d\n", i+1 );
#endif
    }
     
  return;
}

Distribution* ReadDistribution( FILE* input ) {
  /* Reads parameters of the sub-components of the distribution and creates them */

  /* Distribution */
  Distribution* dist;

  /* Parameters */
  long t, param1,param2;
  double mdl_prior;         // Penalty term for parameters
  double pcount;            // Pseudo-count -- Dirichlet prior

  /* Index variable(s) */
  long i,j,b1,b2;

  /* Temporary variable(s) */
  char* temp_char;
  long* dim_index;

  temp_char=read_word( input );
  if( strcmp( temp_char, "factored" )==0 )
    t=DIST_FACTOR;
  else if( strcmp( temp_char, "independent" )==0 )
    t=DIST_ALIAS_CI;
  else if( strcmp( temp_char, "mixture-distinct" )==0 )
    t=DIST_MIXTURE;
  else if( strcmp( temp_char, "mixture-distinct-prior" )==0 )
    t=DIST_MIXTURE_PRIOR;
  else if( strcmp( temp_char, "mixture" )==0 )
    t=DIST_ALIAS_MIXTURE;
  else if( strcmp( temp_char, "mixture-prior" )==0 )
    t=DIST_ALIAS_MIXTURE_PRIOR;
  else if( strcmp( temp_char, "cl-mixture-distinct" )==0 )
    t=DIST_CLMIXTURE;
  else if( strcmp( temp_char, "cl-mixture" )==0 )
    t=DIST_ALIAS_CLMIXTURE;
  else if( strcmp( temp_char, "cl-mixture-prior" )==0 )
    t=DIST_ALIAS_CLMIXTURE_PRIOR;
  else if( strcmp( temp_char, "bernoulli" )==0 )
    t=DIST_BERNOULLI;
  else if( strcmp( temp_char, "bernoulli-prior" )==0 )
    t=DIST_BERNOULLI_PRIOR;
  else if( strcmp( temp_char, "chain-bernoulli" )==0 )
    t=DIST_CONDBERNOULLI;
  else if( strcmp( temp_char, "chain-bernoulli-prior" )==0 )
    t=DIST_CONDBERNOULLI_PRIOR;
  else if( strcmp( temp_char, "chain-bernoulli-global" )==0 )
    t=DIST_CONDBERNOULLIG;
  else if( strcmp( temp_char, "chain-bernoulli-global-prior" )==0 )
    t=DIST_CONDBERNOULLIG_PRIOR;
  else if( strcmp( temp_char, "chow-liu" )==0 )
    t=DIST_CHOWLIU;
  else if( strcmp( temp_char, "chow-liu-mdl" )==0 )
    t=DIST_CHOWLIU_MDL;
  else if( strcmp( temp_char, "chow-liu-prior" )==0 )
    t=DIST_CHOWLIU_DIR_MDL;
  else if( strcmp( temp_char, "conditional-chow-liu" )==0 )
    t=DIST_CONDCHOWLIU;
  else if( strcmp( temp_char, "conditional-chow-liu-mdl" )==0 )
    t=DIST_CONDCHOWLIU_MDL;
  else if( strcmp( temp_char, "maxent-full" )==0 )
    t=DIST_ME_BIVAR;
  else if( strcmp( temp_char, "BN-maxent" )==0 )
    t=DIST_BN_ME_BIVAR;
  else if( strcmp( temp_char, "BN-cond-maxent" )==0 )
    t=DIST_BN_CME_BIVAR;
  else if( strcmp( temp_char, "delta-exponential" )==0 )
    t=DIST_DELTAEXP;
  else if( strcmp( temp_char, "delta-gamma" )==0 )
    t=DIST_DELTAGAMMA;
  else if( strcmp( temp_char, "delta" )==0 )
    t=DIST_DIRACDELTA;
  else if( strcmp( temp_char, "exponential" )==0 )
    t=DIST_EXP;
  else if( strcmp( temp_char, "gamma" )==0 )
    t=DIST_GAMMA;
  else if( strcmp( temp_char, "log-normal" )==0 )
    t=DIST_LOGNORMAL;
  else if( strcmp( temp_char, "gaussian" )==0 )
    t=DIST_NORMAL;
  else if( strcmp( temp_char, "chain-gaussian" )==0 )
    t=DIST_NORMALCHAIN;
  else if( strcmp( temp_char, "tree-gaussian" )==0 )
    t=DIST_NORMALCL;
  else if( strcmp( temp_char, "logistic" )==0 )
    t=DIST_LOGISTIC;
  else if( strcmp( temp_char, "transition-logistic" )==0 )
    t=DIST_TRANSLOGISTIC;
  else
    {
      fprintf( stderr, "Unknown distribution type %s on line %ld.  Skipping.\n", temp_char, line_number );
    }
	      
  delete temp_char;

  switch( t ) {
  case DIST_FACTOR:
    /* Product distribution (Conditional independence) */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the number of conditionally independent distributions must be at least 1.  Aborting.\n",
		 line_number );
	exit( -1 );	
      }

    dist=new Distribution( DIST_FACTOR, 0, param1 );
#ifdef INPUT_VERBOSE
    printf( "%ld-component product distribution\n", dist->dim );
#endif    
    for( i=0; i<dist->dim; i++ )
      dist->subdist[0][i]=ReadDistribution( input );
    dim_index=new long[param1];
    for( i=0; i<param1; i++ )
      dim_index[i]=i;
    dist->UpdateIndex( dim_index );
    delete dim_index;

    break;
  case DIST_ALIAS_CI:
    /* Conditionally independent distributions of the same type (shortcut) */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the number of conditionally independent distributions must be at least 1.  Aborting.\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( DIST_FACTOR, 0, param1 );
#ifdef INPUT_VERBOSE
    printf( "%ld-component product distribution\n", dist->dim );
#endif    
    dist->subdist[0][0]=ReadDistribution( input );
    for( i=1; i<dist->dim; i++ )
      dist->subdist[0][i]=dist->subdist[0][0]->copy();
    dim_index=new long[param1];
    for( i=0; i<param1; i++ )
      dim_index[i]=i;
    dist->UpdateIndex( dim_index );
    delete dim_index;
    
    break;
  case DIST_MIXTURE:
  case DIST_MIXTURE_PRIOR:
    /* Mixture */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the number of components in a mixture model must be at least 1.  Aborting.\n",
		 line_number );
      }
    dist=new Distribution( DIST_BERNOULLI, param1, 1 );
    
    if( t==DIST_MIXTURE_PRIOR )
      { /* Reading in the parameters for the prior */
	pcount=read_double( input );
	if( pcount<0.0 )
	  {
	    fprintf( stderr, "Error on line %ld: the pseudo-counts for Dirichlet prior for mixing probabilities of a mixture must be non-negative.  Setting them to zero.\n",
		     line_number );
	    pcount=0.0;
	  }
	for( i=0; i<dist->num_states; i++ )
	  dist->pcount_single[i]=pcount;
	dist->pcount=pcount*dist->num_states;
      } /* Reading in the parameters for the prior */
#ifdef INPUT_VERBOSE
    printf( "%ld-component mixture distribution via a multinomial distribution\n", dist->num_states );
    if( t==DIST_MIXTURE_PRIOR )
      printf( "Dirichlet prior with pseudo-counts %lf\n", pcount );      
#endif
    
    dist->subdist=new Distribution**[1];
    dist->subdist[0]=new Distribution*[dist->num_states];
    for( i=0; i<dist->num_states; i++ )
      dist->subdist[0][i]=ReadDistribution( input );
    break;
  case DIST_ALIAS_MIXTURE:
  case DIST_ALIAS_MIXTURE_PRIOR:
    /* Mixture of distributions of the same type (shortcut) */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the number of mixture components must be at least 1.  Aborting.\n",
		 line_number );
	exit( -1 );
      }

    dist=new Distribution( DIST_BERNOULLI, param1, 1 );

    if( t==DIST_ALIAS_MIXTURE_PRIOR )
      { /* Reading in the parameters for the prior */
	pcount=read_double( input );
	if( pcount<0.0 )
	  {
	    fprintf( stderr, "Error on line %ld: the pseudo-counts for Dirichlet prior for mixing probabilities of a mixture must be non-negative.  Setting them to zero.\n",
		     line_number );
	    pcount=0.0;
	  }
	for( i=0; i<dist->num_states; i++ )
	  dist->pcount_single[i]=pcount;
	dist->pcount=pcount*dist->num_states;
      } /* Reading in the parameters for the prior */
#ifdef INPUT_VERBOSE
    printf( "%ld-component mixture distribution defined via a multinomial distribution\n", dist->num_states );
    if( t==DIST_ALIAS_MIXTURE_PRIOR )
      printf( "Dirichlet prior with pseudo-counts %lf\n", pcount );      
#endif
    dist->subdist=new Distribution**[1];
    dist->subdist[0]=new Distribution*[dist->num_states];
    dist->subdist[0][0]=ReadDistribution( input );
    for( i=1; i<dist->num_states; i++ )
      dist->subdist[0][i]=dist->subdist[0][0]->copy();
    break;
  case DIST_CLMIXTURE:
    /* Tree-dependent mixture */
    /* First parameter -- number of variables in each output node */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of variables in the Chow-Liu tree must be at least 2.  Aborting\n",
		 line_number );
	exit( -1 );
      }
    
    /* !!! Assuming that each of the variables has the same range !!! */
    /* Second parameter -- number of values for each variable */
    param2=read_long( input );
    if( param2<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of possible values for each variable in the Chow-Liu tree distribution must be at least 2. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    dist=new Distribution( DIST_CHOWLIU, param2, param1 );
    pcount=read_double( input );
    if( pcount<0.0 )
      {
	fprintf( stderr, "Error on line %ld: the pseudo-counts for Dirichlet prior for Chow-Liu tree parameters must be non-negative.  Setting them to zero.\n", line_number );
	pcount=0.0;
      }
    /* !!! No safety checking for the priors !!! */
    for( i=0; i<dist->dim; i++ )
      for( j=0; j<i; j++ )
	for( b1=0; b1<dist->num_states; b1++ )
	  for( b2=0; b2<dist->num_states; b2++ )
	    dist->pcount_bi[i][j][b1][b2]=pcount;
    /* !!! Shortcut for the computation of the marginals !!! */
    for( i=0; i<dist->dim; i++ )
      for( b1=0; b1<dist->num_states; b1++ )
	dist->pcount_uni[i][b1]=pcount*dist->num_states;
    dist->pcount=pcount*dist->num_states*dist->num_states;
    mdl_prior=read_double( input );
    dist->mdl_beta=mdl_prior;
#ifdef INPUT_VERBOSE
    printf( "%ld-variable %ld-component tree-structured mixture with Dirichlet prior with %lf pseudo-counts and MDL prior with factor %lf\n", dist->dim, dist->num_states, pcount, dist->mdl_beta );
#endif
    dist->subdist=new Distribution**[dist->dim];
    for( i=0; i<dist->dim; i++ )
      dist->subdist[i]=new Distribution*[dist->num_states];
    for( i=0; i<dist->dim; i++ )
      for( j=0; j<dist->num_states; j++ )
	dist->subdist[i][j]=ReadDistribution( input );
    dim_index=new long[param1];
    for( i=0; i<param1; i++ )
      dim_index[i]=i;
    dist->UpdateIndex( dim_index );
    delete dim_index;
    break;
  case DIST_ALIAS_CLMIXTURE:
  case DIST_ALIAS_CLMIXTURE_PRIOR:
    /* Tree-structured mixture of distributions of the same type (shortcut) */
    /* First parameter -- number of variables in each output node */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of variables in the Chow-Liu tree must be at least 2.  Aborting\n",
		 line_number );
	exit( -1 );
      }
    
    /* !!! Assuming that each of the variables has the same range !!! */
    /* Second parameter -- number of values for each variable */
    param2=read_long( input );
    if( param2<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of possible values for each variable in the Chow-Liu tree distribution must be at least 2. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    dist=new Distribution( DIST_CHOWLIU, param2, param1 );
    if( t==DIST_ALIAS_CLMIXTURE )
      {
	mdl_prior=read_double( input );
	dist->mdl_beta=mdl_prior;
#ifdef INPUT_VERBOSE
	printf( "%ld-variable %ld-component tree-structured mixture with MDL factor %0.6le\n", 
		dist->dim, dist->num_states, dist->mdl_beta );
#endif
      }
    else if( t==DIST_ALIAS_CLMIXTURE_PRIOR )
      {
	pcount=read_double( input );
	if( pcount<0.0 )
	  {
	    fprintf( stderr, "Error on line %ld: the pseudo-counts for Dirichlet prior for Chow-Liu tree parameters must be non-negative.  Setting them to zero.\n", line_number );
	    pcount=0.0;
	  }
	/* !!! No safety checking for the priors !!! */
	for( i=0; i<dist->dim; i++ )
	  for( j=0; j<i; j++ )
	    for( b1=0; b1<dist->num_states; b1++ )
	      for( b2=0; b2<dist->num_states; b2++ )
		dist->pcount_bi[i][j][b1][b2]=pcount;
	/* !!! Shortcut for the computation of the marginals !!! */
	for( i=0; i<dist->dim; i++ )
	  for( b1=0; b1<dist->num_states; b1++ )
	    dist->pcount_uni[i][b1]=pcount*dist->num_states;
	dist->pcount=pcount*dist->num_states*dist->num_states;
	mdl_prior=read_double( input );
	dist->mdl_beta=mdl_prior;
#ifdef INPUT_VERBOSE
	printf( "%ld-variable %ld-component tree-structured mixture with Dirichlet prior with %lf pseudo-counts and MDL prior with factor %lf\n", dist->dim, dist->num_states, pcount, dist->mdl_beta );
#endif
      }
    dist->subdist=new Distribution**[dist->dim];
    for( i=0; i<dist->dim; i++ )
      dist->subdist[i]=new Distribution*[dist->num_states];
    for( j=0; j<dist->num_states; j++ )
      dist->subdist[0][j]=ReadDistribution( input );
    for( i=1; i<dist->dim; i++ )
      for( j=0; j<dist->num_states; j++ )
	dist->subdist[i][j]=dist->subdist[0][j]->copy();
    dim_index=new long[param1];
    for( i=0; i<param1; i++ )
      dim_index[i]=i;
    dist->UpdateIndex( dim_index );
    delete dim_index;
    break;
  case DIST_BERNOULLI:
  case DIST_BERNOULLI_PRIOR:
    /* One-dimensional Bernoulli */
    
    /* One parameter -- number of possible states */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the number of states for a 1-d Bernoulli distribution must be at least 1.  Aborting.\n",
		 line_number );
	exit( -1 );
      }

    dist=new Distribution( DIST_BERNOULLI, param1, 1 );

    if( t==DIST_BERNOULLI_PRIOR )
      { /* Reading in the parameters for the prior */
	pcount=read_double( input );
	if( pcount<0.0 )
	  {
	    fprintf( stderr, "Error on line %ld: the pseudo-counts for Dirichlet prior of 1-d multinomial distribution must be non-negative.  Setting them to zero.\n",
		     line_number );
	    pcount=0.0;
	  }
	for( i=0; i<dist->num_states; i++ )
	  dist->pcount_single[i]=pcount;
	dist->pcount=0.0;
	for( i=0; i<dist->num_states; i++ )
	  dist->pcount+=dist->pcount_single[i];
      } /* Reading in the parameters for the prior */

    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "Univariate %ld-valued Bernoulli\n", dist->num_states );
    if( t==DIST_BERNOULLI_PRIOR )
      printf( "Dirichlet prior with pseudo-counts %lf\n", pcount );
#endif
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
  case DIST_CONDBERNOULLI_PRIOR:
  case DIST_CONDBERNOULLIG_PRIOR:
    /* Conditional Bernoulli */
    
    /* One parameter -- number of possible states */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the number of states for a conditional Bernoulli distribution must be at least 1. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    
    switch( t ) {
    case DIST_CONDBERNOULLI:
    case DIST_CONDBERNOULLI_PRIOR:
      dist=new Distribution( DIST_CONDBERNOULLI, param1, 1 );
      break;
    case DIST_CONDBERNOULLIG:
    case DIST_CONDBERNOULLIG_PRIOR:
      dist=new Distribution( DIST_CONDBERNOULLIG, param1, 1 );
      break;
    }

    if( t==DIST_CONDBERNOULLI_PRIOR || t==DIST_CONDBERNOULLIG_PRIOR )
      { /* Reading in the parameters for the prior */
	pcount=read_double( input );
	if( pcount<0.0 )
	  {
	    fprintf( stderr, "Error on line %ld: the pseudo-counts for Dirichlet prior of 1-d conditional multinomial distribution must be non-negative.  Setting them to zero.\n",
		     line_number );
	    pcount=0.0;
	  }
	for( i=0; i<dist->num_states; i++ )
	  for( j=0; j<dist->num_states; j++ )
	    dist->pcount_uni[i][j]=pcount;
	for( i=0; i<dist->num_states; i++ )
	  dist->pcount_single[i]=pcount*dist->num_states;
	dist->pcount=pcount*dist->num_states*dist->num_states;
      } /* Reading in the parameters for the prior */

    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    if( t==DIST_CONDBERNOULLI )
      printf( "Univariate %ld-valued conditional Bernoulli\n", dist->num_states );
    else if( t==DIST_CONDBERNOULLIG )
      printf( "Univariate %ld-valued conditional Bernoulli with averaged first sequence value\n", dist->num_states );
    else if( t==DIST_CONDBERNOULLI_PRIOR )
      printf( "Univariate %ld-valued conditional Bernoulli with Dirichlet prior with pseudo-counts %lf\n", dist->num_states, pcount );
    else if( t==DIST_CONDBERNOULLIG_PRIOR )
      printf( "Univariate %ld-valued conditional Bernoulli with averaged first sequence value with Dirichlet prior with pseudo-counts %lf\n", dist->num_states, pcount );      
#endif
    break;
  case DIST_UNICONDMVME:
    /* Conditional multi-valued MaxEnt */

    /* !!! Not to be read in directly !!! */
    break;
  case DIST_CHOWLIU:
  case DIST_CHOWLIU_MDL:
  case DIST_CHOWLIU_DIR_MDL:
    /* Chow-Liu tree */
    
    /* First parameter -- number of variables in each output node */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of variables in the Chow-Liu tree must be at least 2.  Aborting\n",
		 line_number );
	exit( -1 );
      }
    
    /* !!! Assuming that each of the variables has the same range !!! */
    /* Second parameter -- number of values for each variable */
    param2=read_long( input );
    if( param2<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of possible values for each variable in the Chow-Liu tree distribution must be at least 2. Aborting.",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( DIST_CHOWLIU, param2, param1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;

    /* Priors */
    if( t==DIST_CHOWLIU_MDL )
      {
	mdl_prior=read_double( input );
	dist->mdl_beta=mdl_prior;
#ifdef INPUT_VERBOSE
	printf( "%ld-variate %ld-valued Chow-Liu tree with MDL prior with factor %0.6le\n",
		dist->dim, dist->num_states, dist->mdl_beta );
#endif
      }
    else if( t==DIST_CHOWLIU_DIR_MDL )
      {
	pcount=read_double( input );
	if( pcount<0.0 )
	  {
	    fprintf( stderr, "Error on line %ld: the pseudo-counts for Dirichlet prior for Chow-Liu tree parameters must be non-negative.  Setting them to zero.\n", line_number );
	    pcount=0.0;
	  }
	/* !!! No safety checking for the priors !!! */
	for( i=0; i<dist->dim; i++ )
	  for( j=0; j<i; j++ )
	    for( b1=0; b1<dist->num_states; b1++ )
	      for( b2=0; b2<dist->num_states; b2++ )
		dist->pcount_bi[i][j][b1][b2]=pcount;
	/* !!! Shortcut for the computation of the marginals !!! */
	for( i=0; i<dist->dim; i++ )
	  for( b1=0; b1<dist->num_states; b1++ )
	    dist->pcount_uni[i][b1]=pcount*dist->num_states;
	dist->pcount=pcount*dist->num_states*dist->num_states;
	mdl_prior=read_double( input );
	dist->mdl_beta=mdl_prior;
#ifdef INPUT_VERBOSE
	printf( "%ld-variate %ld-valued Chow-Liu tree with Dirichlet prior with %lf pseudo-counts and MDL prior with factor %lf\n", dist->dim, dist->num_states, pcount, dist->mdl_beta );
#endif
      }
    else if( t==DIST_CHOWLIU )
      {
#ifdef INPUT_VERBOSE
	printf( "%ld-variate %ld-valued Chow-Liu tree\n", dist->dim, dist->num_states );
#endif    
	dist->mdl_beta=0.0;
      }
    break;
  case DIST_CONDCHOWLIU:
  case DIST_CONDCHOWLIU_MDL:
    /* Conditional Chow-Liu tree */
    
    /* First parameter -- number of variables in each output node */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of variables in the conditional Chow-Liu tree must be at least 2.  Aborting\n",
		 line_number );
	exit( -1 );
      }
    
    /* !!! Assuming that each of the variables has the same range !!! */
    /* Second parameter -- number of values for each variable */
    param2=read_long( input );
    if( param2<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of possible values for each variable in the conditional Chow-Liu tree distribution must be at least 2. Aborting.",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( DIST_CONDCHOWLIU, param2, param1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
    /* MDL factor */
    if( t==DIST_CONDCHOWLIU_MDL )
      {
	mdl_prior=read_double( input );
	dist->mdl_beta=mdl_prior;
#ifdef INPUT_VERBOSE
	printf( "%ld-variate %ld-valued conditional Chow-Liu tree with MDL prior with factor %0.6le\n",
		dist->dim, dist->num_states, dist->mdl_beta );
#endif
      }
    else if( t==DIST_CONDCHOWLIU )
      {
	dist->mdl_beta=0.0;
#ifdef INPUT_VERBOSE
	printf( "%ld-variate %ld-valued conditional Chow-Liu tree\n", dist->dim, dist->num_states );
#endif
      }
    break;
  case DIST_ME_BIVAR:
    /* Binary bivariate MaxEnt distribution */

    /* First parameter -- number of stations */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of nodes in the full bivariate MaxEnt distribution must be at least 2.  Aborting\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, 2, param1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "%ld-variate binary full bivariate MaxEnt\n", dist->dim );
#endif
    break;
  case DIST_BN_ME_BIVAR:
    /* Bayesian network binary bivariate MaxEnt distribution */

    /* First parameter -- number of stations */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of nodes in the BN bivariate MaxEnt distribution must be at least 2.  Aborting\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, 2, param1 );
    mdl_prior=read_double( input );
    dist->mdl_beta=mdl_prior;
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "%ld-variate binary BN bivariate MaxEnt with MDL factor %0.6le\n", dist->dim, dist->mdl_beta );
#endif
    break;
   case DIST_BN_CME_BIVAR:
    /* Bayesian network binary bivariate conditional MaxEnt distribution */

    /* First parameter -- number of stations */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of nodes in the BN bivariate conditional MaxEnt distribution must be at least 2.  Aborting\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, 2, param1 );
    mdl_prior=read_double( input );
    dist->mdl_beta=mdl_prior;
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "%ld-variate binary BN bivariate conditional MaxEnt with MDL factor %0.6le\n", dist->dim, dist->mdl_beta );
#endif
    break;
 case DIST_DELTAEXP:
    /* Number of components */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of components of the delta-exponential distribution must be at least 2. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, param1, 1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "Univariate %ld-component delta-exponential mixture\n", dist->num_states );
#endif
    break;
  case DIST_DELTAGAMMA:
    /* Number of components */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of components of the delta-gamma distribution must be at least 2. Aborting.",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, param1, 1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "Univariate %ld-component delta-gamma mixture\n", dist->num_states );
#endif
    break;
  case DIST_DIRACDELTA:
    /* Dirac's delta function */

    /* Location of the delta function */
    mdl_prior=read_double( input );

    dist=new Distribution( t, 0, 1 );

    /* Updating the value */
    dist->delta_value=mdl_prior;

    /* By default, the component index is 0 */
    dist->dim_index[0]=0;
#ifdef INPUT_VERBOSE
    printf( "Dirac's delta function at %le\n", dist->delta_value );
#endif
    break;
  case DIST_EXP:
    /* Number of components */
    dist=new Distribution( t, 0, 1 );
    dist->dim_index[0]=0;
#ifdef INPUT_VERBOSE
    printf( "Univariate exponential (geometric) distribution\n" );
#endif
    break;
  case DIST_GAMMA:
    dist=new Distribution( t, 0, 1 );
    dist->dim_index[0]=0;
#ifdef INPUT_VERBOSE
    printf( "Univariate gamma distribution\n" );
#endif
    break;
  case DIST_LOGNORMAL:
    dist=new Distribution( t, 0, 1 );
    dist->dim_index[0]=0;
#ifdef INPUT_VERBOSE
    printf( "Univariate log-normal distribution\n" );
#endif
    break;
  case DIST_NORMAL:
    /* Gaussian distribution */
    
    /* One parameter -- the dimension of the space */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the dimensionality of the Gaussian distribution must be at least 1. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, 0, param1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "%ld-variate Gaussian\n", dist->dim );
#endif
    break;
  case DIST_NORMALCHAIN:
    /* AR-Gaussian distribution */
    
    /* One parameter -- the dimension of the space */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the dimensionality of the AR-Gaussian distribution must be at least 1. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, 0, param1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "%ld-variate AR-Gaussian\n", dist->dim );
#endif
    break;
  case DIST_NORMALCL:
    /* Tree-structured Gaussian distribution */
    
    /* One parameter -- the dimension of the space */
    param1=read_long( input );
    if( param1<1 )
      {
	fprintf( stderr, "Error on line %ld: the dimensionality of the tree-structured Gaussian distribution must be at least 1. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, 0, param1 );
    for( i=0; i<dist->dim; i++ )
      dist->dim_index[i]=i;
#ifdef INPUT_VERBOSE
    printf( "%ld-variate tree-structured Gaussian\n", dist->dim );
#endif
    break;
  case DIST_LOGISTIC:
    /* Logistic function */
    /* Two parameters -- number of states and the dimension of the input space */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of states in a logistic distribution must be at least 2. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    param2=read_long( input );
    if( param2<1 )
      {
	fprintf( stderr, "Error on line %ld: the dimensionality of a logistic distribution must be at least 1. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    
    dist=new Distribution( t, param1, param2 );
#ifdef INPUT_VERBOSE
    printf( "%ld-dimensional %ld-valued logistic\n", dist->dim, dist->num_states );
#endif
    break;
  case DIST_TRANSLOGISTIC:
    /* Trans-logistic function */
    
    /* Two parameters -- number of states and the dimension of the input space */
    param1=read_long( input );
    if( param1<2 )
      {
	fprintf( stderr, "Error on line %ld: the number of states in a trans-logistic distribution must be at least 2. Aborting.\n",
		 line_number );
	exit( -1 );
      }
    param2=read_long( input );
    if( param2<1 )
      {
	    fprintf( stderr, "Error on line %ld: the dimensionality of a trans-logistic distribution must be at least 1. Aborting.\n",
		     line_number );
	    exit( -1 );
      }
    
    dist=new Distribution( t, param1, param2 );
#ifdef INPUT_VERBOSE
    printf( "%ld-dimensional %ld-valued trans-logistic\n", dist->dim, dist->num_states );
#endif
    break;
  default:
    fprintf( stderr, "Error on line %ld: unknown distribution code: %ld. Aborting.\n", line_number, t );
    exit( -1 );
  }

  return( dist );
}

void Distribution::UpdateIndex( long* index ) {
  /* Index variable(s) */
  long i,j;

  switch( type ) {
  case DIST_FACTOR:
    /* Product distribution */
    for( i=0; i<dim; i++ )
      subdist[0][i]->UpdateIndex( &index[i] );
    break;
  case DIST_BERNOULLI:
    if( subdist )
      /* Mixture */
      for( i=0; i<num_states; i++ )
	subdist[0][i]->UpdateIndex( index );
    else
      dim_index[0]=index[0];  
    break;
  case DIST_CHOWLIU:
    if( subdist )
      /* Mixture */
      for( i=0; i<dim; i++ )
	for( j=0; j<num_states; j++ )
	  subdist[i][j]->UpdateIndex( &index[i] );
    else
      for( i=0; i<dim; i++ )
	dim_index[i]=index[i];      
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
  case DIST_DELTAEXP:
  case DIST_DELTAGAMMA:
  case DIST_DIRACDELTA:
  case DIST_EXP:
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
    dim_index[0]=index[0];
    break;
  case DIST_UNICONDMVME:
    break;
  case DIST_CONDCHOWLIU:
  case DIST_ME_BIVAR:
  case DIST_BN_ME_BIVAR:
  case DIST_BN_CME_BIVAR:
  case DIST_NORMAL:
  case DIST_NORMALCHAIN:
  case DIST_NORMALCL:
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
    for( i=0; i<dim; i++ )
      dim_index[i]=index[i];
    break;
  default:;
  }

  return;
}

void Distribution::ReadParameters( FILE* input ) {
  /* Reads the values of parameters */

  /* Index variable(s) */
  long i,j,s;

  switch( type ) {
  case DIST_FACTOR:
    /* Reading parameters for sub-components */
    for( i=0; i<dim; i++ )
      subdist[0][i]->ReadParameters( input );
    break;
  case DIST_BERNOULLI:
    /* Initializing Bernoulli parameters */
    if( !subdist )
      for( i=0; i<dim; i++ )
	dim_index[i]=read_long( input );
    for( i=0; i<num_states; i++ )
      state_prob[i]=read_double( input );
    for( i=0; i<num_states; i++ )
      log_state_prob[i]=log(state_prob[i]);
    if( subdist )
      /* Reading mixture components */
      for( i=0; i<dim; i++ )
	for( j=0; j<num_states; j++ )
	  subdist[i][j]->ReadParameters( input );
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* Conditional Bernoulli distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );

    /* First sequence entry probabilities */
    for( i=0; i<num_states; i++ )
      state_prob[i]=read_double( input );
    for( i=0; i<num_states; i++ )
      log_state_prob[i]=log(state_prob[i]);

    /* Conditional probabilities */
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	cond_state_prob[i][j]=read_double( input );
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	log_cond_state_prob[i][j]=log(cond_state_prob[i][j]);
    break;
  case DIST_UNICONDMVME:
    /* Multinomial MaxEnt distribution */
    
    /* Number of functions */
    dim=read_long( input );

    /* Number of features for each function */
    for( i=0; i<dim; i++ )
      num_features[i]=read_long( input );

    /* Indices of features for functions */
    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_index[i][j]=read_long( input );

    /* Corresponding feature values */
    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_value[i][j]=read_long( input );

    /* Weights */
    for( i=0; i<dim; i++ )
      lambda[i]=read_double( input );
    break;
  case DIST_CHOWLIU:
    /* Chow-Liu tree */
    if( !subdist )
      /* Not a mixture */
      for( i=0; i<dim; i++ )
	dim_index[i]=read_long( input );

    /* Marginal probabilities */
    for( i=0; i<dim; i++ )
      for( s=0; s<num_states; s++ )
	ind_prob[i][s]=read_double( input );

    /* Number of edges */
    num_edges=read_long( input );

    /* Nodes on the edges */
    for( i=0; i<num_edges; i++ )
      {
	edge[i][0]=read_long( input );
	edge[i][1]=read_long( input );
      }

    /* Probabilities on the edges */
    for( i=0; i<num_edges; i++ )
      for( j=0; j<num_states; j++ )
	for( s=0; s<num_states; s++ )
	  edge_prob[i][j][s]=read_double( input );

    if( subdist )
      { /* Mixture */
	/* Reading mixing components */
	for( i=0; i<dim; i++ )
	  for( j=0; j<num_states; j++ )
	    subdist[i][j]->ReadParameters( input );
	
	/* Computing the envelopes */
	for( i=0; i<num_envelopes; i++ )
	  delete envelope[i];
	
	num_envelopes=compute_envelopes_full();    
     } /* Mixture */

    break;
  case DIST_CONDCHOWLIU:
    /* Conditional Chow-Liu tree */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );

    /* Marginal probabilities */
    for( i=0; i<2*dim; i++ )
      for( s=0; s<num_states; s++ )
	ind_prob[i][s]=read_double( input );

    /* Number of edges */
    num_edges=read_long( input );

    /* Nodes on the edges */
    for( i=0; i<num_edges; i++ )
      {
	edge[i][0]=read_long( input );
	edge[i][1]=read_long( input );
      }

    /* Probabilities on the edges */
    for( i=0; i<num_edges; i++ )
      for( j=0; j<num_states; j++ )
	for( s=0; s<num_states; s++ )
	  edge_prob[i][j][s]=read_double( input );

    break;
  case DIST_ME_BIVAR:
    /* Binary full bivariate MaxEnt distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );

    /* Coefficients */
    for( i=0; i<dim; i++ )
      for( j=0; j<i+1; j++ )
	sigma[i][j]=read_double( input );

    /* Normalization constant exponent */
    det=read_double( input );
    break;
  case DIST_BN_ME_BIVAR:
    /* Bayesian network binary bivariate MaxEnt distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );
    
    for( i=0; i<dim; i++ )
      num_features[i]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_index[i][j]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	sigma[i][j]=read_double( input );


    for( i=0; i<dim; i++ )
      sim_order[i]=read_long( input );
    break;
  case DIST_BN_CME_BIVAR:
    /* Bayesian network binary bivariate MaxEnt conditional distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );
    
    for( i=0; i<dim; i++ )
      state_prob[i]=read_double( input );

    for( i=0; i<dim; i++ )
      num_features[i]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_index[i][j]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	sigma[i][j]=read_double( input );


    for( i=0; i<dim; i++ )
      sim_order[i]=read_long( input );
    break;
  case DIST_DELTAEXP:
    /* Delta-exponential distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );
    
    /* Reading mixing probabilities */
    for( i=0; i<num_states; i++ )
      mix_prob[i]=read_double( input );

    /* Reading the parameters for the exponential components */
    for( i=0; i<num_states-1; i++ )
      exp_param[i]=read_double( input );

    break;
  case DIST_DELTAGAMMA:
    /* Delta-gamma distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );
    
    /* Reading mixing probabilities */
    for( i=0; i<num_states; i++ )
      mix_prob[i]=read_double( input );

    /* Reading the parameters for the gamma components */
    for( i=0; i<num_states-1; i++ )
      {
	gamma_param1[i]=read_double( input );
	gamma_param2[i]=read_double( input );
      }
    break;
  case DIST_DIRACDELTA:
    /* Univariate Dirac delta */
    dim_index[0]=read_long( input );
    break;
  case DIST_EXP:
    /* Univariate exponential distribution */
    dim_index[0]=read_long( input );
    
    /* Reading the parameter */
    exp_param1=read_double( input );

    break;
  case DIST_GAMMA:
    /* Univariate gamma distribution */
    dim_index[0]=read_long( input );
    
    /* Reading in two parameters */
    gamma1=read_double( input );
    gamma2=read_double( input );
    break;
  case DIST_LOGNORMAL:
    /* Univariate gamma distribution */
    dim_index[0]=read_long( input );
    
    /* Reading in two parameters */
    log_normal1=read_double( input );
    log_normal2=read_double( input );
    break;
  case DIST_NORMAL:
    /* Normal distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );

    /* Reading in the mean vector */
    for( i=0; i<dim; i++ )
      mu[i]=read_double( input );
    
    /* Reading in the covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	sigma[i][j]=read_double( input );
    
    /* Need to find the inverse and the determinant */
    find_inv( sigma, dim, inv_sigma );
    det=find_det( sigma, dim );
    break;
  case DIST_NORMALCHAIN:
    /* Normal distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );

    /* Reading in the first state mean vector */
    for( i=0; i<dim; i++ )
      first_mu[i]=read_double( input );
    
    /* Reading in the first state covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	first_sigma[i][j]=read_double( input );
    
    /* Need to find the inverse and the determinant */
    find_inv( first_sigma, dim, inv_first_sigma );
    first_det=find_det( first_sigma, dim );

    /* Reading in the transformation matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	W[i][j]=read_double( input );

    /* Reading in translation vector */
    for( i=0; i<dim; i++ )
      mu[i]=read_double( input );

    /* Reading in noise covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	sigma[i][j]=read_double( input );

    /* Inverting */
    find_inv( sigma, dim, inv_sigma );
    det=find_det( sigma, dim );

    break;
  case DIST_NORMALCL:
    /* Normal distribution */
    for( i=0; i<dim; i++ )
      dim_index[i]=read_long( input );

    /* Reading in the mean vector */
    for( i=0; i<dim; i++ )
      mu[i]=read_double( input );
    
    /* Reading in the covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	sigma[i][j]=read_double( input );
    
    /* Need to find the inverse and the determinant */
    find_inv( sigma, dim, inv_sigma );
    det=find_det( sigma, dim );

    /* Number of edges */
    num_edges=read_long( input );

    /* Nodes on the edges */
    for( i=0; i<num_edges; i++ )
      {
	edge[i][0]=read_long( input );
	edge[i][1]=read_long( input );
      }
    break;
  case DIST_LOGISTIC:
    /* Logistic distribution */
    for( s=0; s<num_states; s++ )
      { /* Parameters by state */
	/* Coefficients for the constant term */
	lambda[s]=read_double( input );

	/* Coefficients for the linear term */
	for( i=0; i<dim; i++ )
	  rho[s][i]=read_double( input );
      } /* Parameters by state */
    break;
  case DIST_TRANSLOGISTIC:
    /* Logistic distribution with transitions */
    for( s=0; s<num_states; s++ )
      { /* Parameters by state */
	/* Coefficients for the constant term */
	lambda[s]=read_double( input );

	/* Coefficients for the transition term */
	for( i=0; i<num_states; i++ )
	  sigma[i][s]=read_double( input );

	/* Coefficients for the linear term */
	for( i=0; i<dim; i++ )
	  rho[s][i]=read_double( input );
      } /* Parameters by state */
    break;
  default:;
  } 
  return;
}

void Distribution::ReadParameters2( FILE* input ) {
  /* Reads the values of parameters */

  /* Not reading in the dimension index */

  /* Index variable(s) */
  long i,j,s;

  switch( type ) {
  case DIST_FACTOR:
    /* Reading parameters for sub-components */
    for( i=0; i<dim; i++ )
      subdist[0][i]->ReadParameters2( input );
    break;
  case DIST_BERNOULLI:
    /* Initializing Bernoulli parameters */
    for( i=0; i<num_states; i++ )
      state_prob[i]=read_double( input );
    for( i=0; i<num_states; i++ )
      log_state_prob[i]=log(state_prob[i]);
    if( subdist )
      /* Mixture components */
      for( i=0; i<dim; i++ )
	for( j=0; j<num_states; j++ )
	  subdist[i][j]->ReadParameters2( input );
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* Conditional Bernoulli distribution */
    /* First sequence entry probabilities */
    for( i=0; i<num_states; i++ )
      state_prob[i]=read_double( input );
    for( i=0; i<num_states; i++ )
      log_state_prob[i]=log(state_prob[i]);

    /* Conditional probabilities */
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	cond_state_prob[i][j]=read_double( input );
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	log_cond_state_prob[i][j]=log(cond_state_prob[i][j]);
    break;
  case DIST_UNICONDMVME:
    /* Multinomial MaxEnt distribution */
    
    /* Number of functions */
    dim=read_long( input );

    /* Number of features for each function */
    for( i=0; i<dim; i++ )
      num_features[i]=read_long( input );

    /* Indices of features for functions */
    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_index[i][j]=read_long( input );

    /* Corresponding feature values */
    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_value[i][j]=read_long( input );

    /* Weights */
    for( i=0; i<dim; i++ )
      lambda[i]=read_double( input );
    break;
  case DIST_CHOWLIU:
    /* Chow-Liu tree */
    /* Marginal probabilities */
    for( i=0; i<dim; i++ )
      for( s=0; s<num_states; s++ )
	ind_prob[i][s]=read_double( input );

    /* Number of edges */
    num_edges=read_long( input );

    /* Nodes on the edges */
    for( i=0; i<num_edges; i++ )
      {
	edge[i][0]=read_long( input );
	edge[i][1]=read_long( input );
      }

    /* Probabilities on the edges */
    for( i=0; i<num_edges; i++ )
      for( j=0; j<num_states; j++ )
	for( s=0; s<num_states; s++ )
	  edge_prob[i][j][s]=read_double( input );

    if( subdist )
      { /* Mixture */
	/* Reading mixing components */
	for( i=0; i<dim; i++ )
	  for( j=0; j<num_states; j++ )
	    subdist[i][j]->ReadParameters2( input );	

	/* Computing the envelopes */
	for( i=0; i<num_envelopes; i++ )
	  delete envelope[i];
	
	num_envelopes=compute_envelopes_full();    
      } /* Mixture */

    break;
  case DIST_CONDCHOWLIU:
    /* Conditional Chow-Liu tree */
    /* Marginal probabilities */
    for( i=0; i<2*dim; i++ )
      for( s=0; s<num_states; s++ )
	ind_prob[i][s]=read_double( input );

    /* Number of edges */
    num_edges=read_long( input );

    /* Nodes on the edges */
    for( i=0; i<num_edges; i++ )
      {
	edge[i][0]=read_long( input );
	edge[i][1]=read_long( input );
      }

    /* Probabilities on the edges */
    for( i=0; i<num_edges; i++ )
      for( j=0; j<num_states; j++ )
	for( s=0; s<num_states; s++ )
	  edge_prob[i][j][s]=read_double( input );

    break;
  case DIST_ME_BIVAR:
    /* Binary full bivariate MaxEnt distribution */
    /* Coefficients */
    for( i=0; i<dim; i++ )
      for( j=0; j<i+1; j++ )
	sigma[i][j]=read_double( input );

    /* Normalization constant exponent */
    det=read_double( input );
    break;
  case DIST_BN_ME_BIVAR:
    /* Bayesian network binary bivariate MaxEnt distribution */
    for( i=0; i<dim; i++ )
      num_features[i]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_index[i][j]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	sigma[i][j]=read_double( input );


    for( i=0; i<dim; i++ )
      sim_order[i]=read_long( input );
    break;
  case DIST_BN_CME_BIVAR:
    /* Bayesian network binary bivariate MaxEnt conditional distribution */
    for( i=0; i<dim; i++ )
      state_prob[i]=read_double( input );

    for( i=0; i<dim; i++ )
      num_features[i]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	feature_index[i][j]=read_long( input );

    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	sigma[i][j]=read_double( input );


    for( i=0; i<dim; i++ )
      sim_order[i]=read_long( input );
    break;
  case DIST_DELTAEXP:
    /* Delta-exponential distribution */
    /* Reading mixing probabilities */
    for( i=0; i<num_states; i++ )
      mix_prob[i]=read_double( input );

    /* Reading the parameters for the exponential components */
    for( i=0; i<num_states-1; i++ )
      exp_param[i]=read_double( input );

    break;
  case DIST_DELTAGAMMA:
    /* Delta-gamma distribution */
    /* Reading mixing probabilities */
    for( i=0; i<num_states; i++ )
      mix_prob[i]=read_double( input );

    /* Reading the parameters for the gamma components */
    for( i=0; i<num_states-1; i++ )
      {
	gamma_param1[i]=read_double( input );
	gamma_param2[i]=read_double( input );
      }
    break;
  case DIST_DIRACDELTA:
    break;
  case DIST_EXP:
    /* Univariate exponential distribution */
    /* Reading the parameter */
    exp_param1=read_double( input );

    break;
  case DIST_GAMMA:
    /* Univariate gamma distribution */
    /* Reading in two parameters */
    gamma1=read_double( input );
    gamma2=read_double( input );
    break;
  case DIST_LOGNORMAL:
    /* Univariate gamma distribution */
    /* Reading in two parameters */
    log_normal1=read_double( input );
    log_normal2=read_double( input );
    break;
  case DIST_NORMAL:
    /* Normal distribution */
    /* Reading in the mean vector */
    for( i=0; i<dim; i++ )
      mu[i]=read_double( input );
    
    /* Reading in the covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	sigma[i][j]=read_double( input );
    
    /* Need to find the inverse and the determinant */
    find_inv( sigma, dim, inv_sigma );
    det=find_det( sigma, dim );
    break;
  case DIST_NORMALCHAIN:
    /* Normal distribution */
    /* Reading in the first state mean vector */
    for( i=0; i<dim; i++ )
      first_mu[i]=read_double( input );
    
    /* Reading in the first state covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	first_sigma[i][j]=read_double( input );
    
    /* Need to find the inverse and the determinant */
    find_inv( first_sigma, dim, inv_first_sigma );
    first_det=find_det( first_sigma, dim );

    /* Reading in the transformation matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	W[i][j]=read_double( input );

    /* Reading in translation vector */
    for( i=0; i<dim; i++ )
      mu[i]=read_double( input );

    /* Reading in noise covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	sigma[i][j]=read_double( input );

    /* Inverting */
    find_inv( sigma, dim, inv_sigma );
    det=find_det( sigma, dim );

    break;
  case DIST_NORMALCL:
    /* Normal distribution with tree-structured covariance matrix */
    /* Reading in the mean vector */
    for( i=0; i<dim; i++ )
      mu[i]=read_double( input );
    
    /* Reading in the covariance matrix */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	sigma[i][j]=read_double( input );
    
    /* Need to find the inverse and the determinant */
    find_inv( sigma, dim, inv_sigma );
    det=find_det( sigma, dim );

    /* Number of edges */
    num_edges=read_long( input );

    /* Nodes on the edges */
    for( i=0; i<num_edges; i++ )
      {
	edge[i][0]=read_long( input );
	edge[i][1]=read_long( input );
      }
    break;
  case DIST_LOGISTIC:
    /* Logistic distribution */
    for( s=0; s<num_states; s++ )
      { /* Parameters by state */
	/* Coefficients for the constant term */
	lambda[s]=read_double( input );

	/* Coefficients for the linear term */
	for( i=0; i<dim; i++ )
	  rho[s][i]=read_double( input );
      } /* Parameters by state */
    break;
  case DIST_TRANSLOGISTIC:
    /* Logistic distribution with transitions */
    for( s=0; s<num_states; s++ )
      { /* Parameters by state */
	/* Coefficients for the constant term */
	lambda[s]=read_double( input );

	/* Coefficients for the transition term */
	for( i=0; i<num_states; i++ )
	  sigma[i][s]=read_double( input );

	/* Coefficients for the linear term */
	for( i=0; i<dim; i++ )
	  rho[s][i]=read_double( input );
      } /* Parameters by state */
    break;
  default:;
  } 
  return;
}

void HMM::ReadParameters( FILE* input_file ) {
  /* Index variable(s) */
  long i,j;

  line_number=1;
  EOF_TRUE=0;

  /* Reading in the transition distribution */  
  transition->ReadParameters2( input_file );
#ifdef INPUT_VERBOSE
  printf( "Transition distribution parameters read\n" );
#endif

  /* Reading in the emissions */
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states; j++ )
      if( !short_dist_display )
	emission[i][j]->ReadParameters( input_file );
      else
	emission[i][j]->ReadParameters2( input_file );
#ifdef INPUT_VERBOSE
  printf( "Emission distribution parameters read\n" );
#endif
  
  return;
}

long read_long( FILE* input ) {
  long output_long;
  char output[100];
  double temp;

  long pos=0;

  eat_blanks( input );
  if( !EOF_TRUE )
    while( is_number_char( output[pos]=fgetc( input ) ) )
      if( pos<100-1 )
	pos++;
      else
	fprintf( stderr, "Number is too long on line %ld\n", line_number );  
  
  ungetc( output[pos], input );
  output[pos]='\0';

  temp=atof( output );
  if( is_missing( temp ) )
    output_long=missing_value( (long)0 );
  else
    output_long=atol( output );
  
  if( errno==EINVAL )
    {
      fprintf( stderr, "Line %ld: problem converting input string %s into long\n", line_number, output );
      exit( -2 );
    }

  return( output_long );
};

double read_double( FILE* input ) {
  double output_double;
  char output[100];

  long pos=0;

  eat_blanks( input );
  if( !EOF_TRUE )
    while( is_number_char( output[pos]=fgetc( input ) ) )
      if( pos<100-1 )
	pos++;
      else
	fprintf( stderr, "String is too long on line %ld\n", line_number );  
  
  if( !EOF_TRUE )
    ungetc( output[pos], input );

  output[pos]='\0';


  output_double=atof( output );
  
  if( errno==EINVAL )
    {
      fprintf( stderr, "Line %ld: problem converting input string %s into double\n", line_number, output );
      exit( -2 );
    }

  return( output_double );
};

long is_number_char( char c ) {
  return( isdigit(c) || c=='e' || c=='E' || c=='.' || c=='+' || c=='-' 
	  || c=='N' || c=='n' || c=='a' || c=='A' );
}

char* read_word( FILE* input ) {
  const int totalChars = 256;
  char* output=new char[totalChars];
  long pos=0;

  eat_blanks( input );
  if( !EOF_TRUE )
    {
      while( is_word_char( output[pos]=fgetc( input ) ) )
	if( pos<totalChars-1 )
	  pos++;
	else
	  fprintf( stderr, "String is too long on line %ld\n", line_number );  
      
      ungetc( output[pos], input );
      output[pos]='\0';
    }

  return( output );
}

#if defined WIN32 || CYGWIN
// On Windows, file name contains 'space' very often. is_word_char is modified to fit this need.
long is_word_char( char c ) {
  
  static bool doubleQuote = false;
  if(c=='\"') 
	  doubleQuote = !doubleQuote;
  if(doubleQuote) 
	  return (c=='\"' || c==' ' || isalnum(c) || c=='_' || c=='/' || c==92 || c==':' || c=='.' || c=='-' );

  return( isalnum(c) || c=='_' || c=='/' || c==92 || c==':' || c=='.' || c=='-' || c=='\"' );
}
#else
long is_word_char( char c ) {
  return( isalnum(c) || c=='_' || c=='/' || c==92 || c==':' || c=='.' || c=='-' );
}
#endif

void eat_blanks( FILE* input ) {
  char c;
  
  do
    {
      c=fgetc( input );
      if( feof( input ) )
	{
	  EOF_TRUE=1;
	  return;
	}
      else
	{
	  switch( c ) {
	  case '\n':
	    line_number++;
	    break;
	    //	  case EOF:
	    //EOF_TRUE=1;
	    //break;
	  case COMMENT_SYMBOL:
	    eat_line( input );
	    c=' ';
	    break;
	  default:
	    ;
	  }
	}
    }
  while( isspace(c) );
  ungetc( c, input );

  return;
}

void eat_line( FILE* input ) {

  while( fgetc(input)!='\n' )
    if( feof( input ) )
      {
	EOF_TRUE=1;
	return;
      }

  line_number++;

  return;
}
