#ifndef MVNHMM_STRUCT_H
#define MVNHMM_STRUCT_H
/*
 * Copyright (c) 2002-2006, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
 */

/* struct.h */

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

/* Flags */
//#define INPUT_VERBOSE
//#define FINAL_CHECK
//#define NEWTONRAPHSON_VERBOSE
//#define HOLE_VERBOSE
//#define SIM_VERBOSE
//#define VITERBI_VERBOSE
//#define XVAL_VERBOSE
//#define VERBOSE_BRIEF
//#define MEASURE_TIME

#ifdef PLATFORM_SUN
#include <ieeefp.h>
#include <sunmath.h>
#endif

#ifdef MEASURE_TIME
#include <time.h>
#endif

#ifdef WIN32
extern "C" {
#include "winExtra.h"
}
#endif

/* Constants */

/* Maximum number of hidden states */
//#define MAXNUMSTATES 20
#define MAXNUMVARIABLES 20      // Ceiling on the number of latent variables (in the Markov Chain) 

/* Data types */
#define DATA_TYPE_FINITE 0
#define DATA_TYPE_REAL   1

/* Model types */
#define MODEL_TYPE_HMM   0      // Hidden Markov model
#define MODEL_TYPE_NHMM  1      // Non-homogeneous hidden Markov model
#define MODEL_TYPE_MIX   2      // Mixture model
#define MODEL_TYPE_NMIX  3      // Non-homogeneous mixture model

/* Distribution type codes */

/* Complex */
#define DIST_FACTOR          101 // Product of (conditionally) independent distributions
#define DIST_MIXTURE         102 // Mixture of distributions
#define DIST_CLMIXTURE       103 // Mixture with tree dependence

/* Univariate finite-valued */
#define DIST_BERNOULLI       1   // Bernoulli distribution
#define DIST_CONDBERNOULLI   2   // Conditional Bernoulli distribution
#define DIST_CONDBERNOULLIG  3   // Same as before but initial sequence value is computed by 
                                 // averaging all entries
#define DIST_UNICONDMVME     4   // Univariate conditional multi-values MaxEnt model

/* Multivariate finite-valued */
#define DIST_CHOWLIU         11  // Distribution with a Chow-Liu tree structure
#define DIST_CONDCHOWLIU     12  // Distribution with a conditional Chow-Liu tree structure
#define DIST_ME_BIVAR        13  // Full bivariate MaxEnt
#define DIST_BN_ME_BIVAR     14  // Bayesian network decomposition using binary uni- and bi-variate MaxEnt 
#define DIST_BN_CME_BIVAR    15  // Same as previous but for conditional distributions

/* Univariate real-valued */
#define DIST_DELTAEXP        21  // Delta-exponential mixture
#define DIST_DELTAGAMMA      22  // Delta-gamma mixture
#define DIST_DIRACDELTA      23  // Dirac delta function
#define DIST_EXP             24  // Exponential (geometric) distribution
#define DIST_GAMMA           25  // Gamma distribution
#define DIST_LOGNORMAL       26  // Log-normal distribution

/* Multivariate real-valued */
#define DIST_NORMAL          31  // Gaussian distribution
#define DIST_NORMALCHAIN     32  // Gaussian dependent on the previous value
#define DIST_NORMALCL        33  // Tree-structured Gaussian

/* Logistic */
#define DIST_LOGISTIC        41  // Logistic distribution
#define DIST_TRANSLOGISTIC   42  // Logistic distribution with transition terms
#define DIST_TRANSLOGISTICG  43  // Same as TRANSLOGISTIC but initial sequence value is computed by
                                 // averaging over all entries

/* Aliases */
#define DIST_ALIAS_CI        51  // Conditionally independent distributions
#define DIST_ALIAS_MIXTURE   52  // Mixture distribution
#define DIST_ALIAS_CLMIXTURE 53  // Tree-dependent mixture distribution
#define DIST_CHOWLIU_MDL     61  // Distribution with a Chow-Liu forest with MDL prior
#define DIST_CONDCHOWLIU_MDL 62  // Distribution with a conditional Chow-Liu forests with MDL prior


/* Cross-validation types */
#define XVAL_NONE      0
#define XVAL_LEAVENOUT 1

/* Analysis types */
#define ANALYSIS_MEAN  0
#define ANALYSIS_CORR  1
#define ANALYSIS_PERS  2
#define ANALYSIS_DRY   3
#define ANALYSIS_WET   4
#define ANALYSIS_MI    5
#define ANALYSIS_LO    6

/* Negative infinity for all practical purposes */
#define NEG_INF -pow(10.0,100000.0)
#define POS_INF pow(10.0,100000.0)

/* Sensitivity constant for EM */
#define EM_EPSILON     5E-05
#define EM_MIX_EPSILON 5E-05

/* Newton-Raphson constants */
#define NR_EPSILON 1E-04
#define MIN_NR_EPSILON 3E-03
#define MAX_NR_ITERATIONS 50

#define INIT_EPSILON    1E-08

/* Comparison epsilon */
#define COMP_EPSILON 1E-12

/* Minimal accepted MaxEnt distance */
#define MAXENT_EPSILON 1E-04

/* Minimal change in log-likelihood needed to add an edge */
#define MIN_ME_LL_CHANGE 1E-02
#define CONJ_GRAD_EPSILON 1E-05

/* Making sure no probability falls out of (1e-12, 1-1e-12) range */
//#define MAX_ABS_PARAM_VALUE log(1E+12-1)

/* Number of burn-off iterations */
#define NUM_BURNOFF1_ITER 5
#define NUM_BURNOFF2_ITER 25
#define NUM_BURNOFF3_ITER 60

/* Maximum number of best sequences for Viterbi */
#define MAXBESTSEQ     1000

/* Maximum number of allowed discrete values per variable (not enforced) */
#define MAXNUMVALUES 100

/* Comment symbol for the parameter (and data) files */
#define COMMENT_SYMBOL '#'

/* Missing value constant for categorical data */
#define CONST_MISSING_VALUE -999

/* Maximum number of features for MaxEnt distributions */
#define MAXNUMMAXENTFEATURES 100

/* Action codes */
#define ACTION_CODE_UNKNOWN   0
#define ACTION_CODE_LEARN     1
#define ACTION_CODE_VITERBI   2
#define ACTION_CODE_LL        3
#define ACTION_CODE_SIM       4
#define ACTION_CODE_ANALYZE   5
#define ACTION_CODE_FILLING   6
#define ACTION_CODE_SIM_FILL  7
#define ACTION_CODE_PREDICT   8
#define ACTION_CODE_DEBUG     1000

/* Types for the coefficients of logistic distribution */
#define PARAM_TYPE_TRANS  0
#define PARAM_TYPE_LINEAR 1

/* Hole-poking evaluation type */
#define POKING_TYPE_LOGP  0
#define POKING_TYPE_PRED  1
#define MISSING_IND_PROB  2
#define HIDDEN_STATE_PROB 3

/* log(2) */
#define CONST_LN2 log(2.0)

/* Initialization type */
#define INIT_RANDOM 0
#define INIT_EM     1

/* Minimum and maximum values for the constraints */
#define CONSTRAINT_MIN_VALUE 1E-012
#define CONSTRAINT_MAX_VALUE 1-1E-012

/* Gibbs sampler parameters */
#define GIBBS_BURNIN 500
#define GIBBS_LAG 30
#define GIBBS_SAMPLES_PER_STATE 1000

/* Parameter for the exponent of the MDL prior */
#define DEFAULT_MDL_BETA 2E-00

/* Classes */
class Parameters {
  /* Contains possible parameters to the program which can be obtained from
     the parameter file.
  */
public:
  /* Read parameters */
  long num_variables;            // Number of latent variables
  long* num_states;              // Number of states for each latent variable
  long model_type;               // Type of the model
  long action;                   // Action for the program to take
  class Distribution** emission; // Emission distributions
  char* output_data_filename;    // Name of the file with the data for the outputs
  long num_ddata_components;     // Number of finite-valued components of the output data
  long num_rdata_components;     // Number of real-valued components of the output data
  long num_data_seqs;            // Number of data sequences
  long length_data_seqs;         // Number of entries per data sequence !!! CAN BE CHANGED LATER !!!
  long input_dim;                // Number of input components
  char* input_filename;          // Name of the file with the data for the inputs
  long num_dinput_components;    // Number of finite-valued components of the input data
  long num_rinput_components;    // Number of real-valued components of the input data
  long num_input_seqs;           // Number of input data sequences
  long length_input_seqs;        // Number of entries per input sequence !!! CAN BE CHANGED LATER !!!
  long xval_type;                // Type of crossvalidation
  long number_out;               // Number of examples to leave out
  char* model_filename;          // Name of the file with the model
  long num_models;               // Number of models in the model file
  char* output_filename;         // Name of the output file
  char* extra_data_filename;     // Name of the extra file (e.g., OLR data)
  long num_dextra_components;    // Number of finite-valued components of the input data
  long num_rextra_components;    // Number of real-valued components of the input data
  long num_extra_seqs;           // Number of input data sequences
  long length_extra_seqs;        // Number of entries per input sequence !!! CAN BE CHANGED LATER !!!
  long num_restarts;             // Number of random restarts of EM
  double em_precision;           // Precision of log-likelihood in EM
  long initialization_type;      // Simple (random) or complex (EM from a simpler model) initialization
  long robust_first_state;       // Probabilities for first state in the sequence are averages
  long num_simulations;          // Number of simulations per
  char* state_filename;          // Name of the file to store the states for simulation
  long analysis_type;            // Type of data analysis
  long poking_type;              // Type of hole-poking analysis
  double cg_epsilon;             // Minimum conjugate gradient change
  double maxent_epsilon;         // Minimum maxent log-likelihood change
  long lookahead;                // Look-ahead

  /* Created parameters */
  class DataPoint* output_datum; // Template for an output data point
  class DataPoint* input_datum;  // Template for an input data point
  class DataPoint* extra_datum;  // Template for an extra data point

  Parameters();
  ~Parameters();
  void Read( FILE* );
};

class HMM {
  /* Class for parameters of the HMM model */
public:
  long num_variables;             // Number of latent variables
  long* num_states;               // Number of states for each variable
  class Distribution* transition; // Transition distribution
  class Distribution*** emission; // Emission distribution
  double ll;                      // Log-likelihood
  double lp;                      // Log-posterior
  double lp_normalized;           // Log-posterior per dimension of data
  double bic;                     // BIC score

  /* EM related */
  double**** pb;                 // Probability values P(R_nt|S_nt=i)
  double** norm_const;           // Normalization factors

  HMM( long, long* );
  ~HMM();
  HMM* copy();
  void Initialize( class Data*, class Data* );
  void InitializeByEM( class Data*, class Data* );
  double log_likelihood( class Data*, class Data* );
  double log_posterior( class Data*, class Data* );
  double hole_score( class Data*, class Data*, long );
  double hole_score_sequence( class Sequence*, class Sequence*, long, long ); 
  void ReadParameters( FILE* );
  void EM( class Data*, class Data*, double );
  void AllocateEMStructures( class Data*, class Data* );
  void AllocateAlpha( class Data*, class Data* );
  void EStep( class Data*, class Data* );
  double log_likelihood_sequence( class Sequence*, class Sequence*, long, long );
  void CalculateProbEmissionsSequence( class Sequence*, long, long );
  void CalculateSummaries( class Data* );
  void DeallocateEMStructures( class Data*, class Data* );
  void DeallocateAlpha( class Data*, class Data* );  

  void CalculateProbEmissions( class Data* );
  void CalculateAlpha( class Data*, class Data* );
  void CalculateBeta( class Data* );

  class StateData* viterbi( class Data*, class Data*, long );
  long** simulate_states( class Data*, long, long );
  long* simulate_state_sequence_gibbs( class Sequence*, class Sequence*, long, long );
  long* simulate_missing_discrete_sequence_gibbs( class Sequence*, long* );
  long*** simulate_latent_sequence_po( class Data*, class Data*, long );
  void InitializeMissingEntriesSequence( class Sequence* );

  void MissingIndividualMarginals( class Data*, class Data*, FILE* );
  void StateProbabilities( class Data*, class Data*, FILE* );

  void WriteToFile( FILE* );

  long num_params();
  double BIC( long );

  void PostProcess();
};

class Distribution {
  /* Class for a distribution */
public:
  long type;                // Type of distribution
  long dim;                 // Dimension of input or the number of subcomponents
  long num_states;          // Number of states for discrete distributions
  long* dim_index;          // Indices of components covered by the distribution
  long num_dists;           // For nested-divide distribution
  double* state_prob;       // Array of probabilities for Bernoulli
  double* mu;               // Means of the gaussian
  double** sigma;           // Covariance matrix or logistic transition matrix
  double** inv_sigma;       // Inverse of the covariance matrix
  double** chol_sigma;      // Upper-triangular Cholesky decomposition 
  double det;               // Determinant of the covariance matrix
  Distribution** subdist;   // Array of sub-components
  Distribution*** subdist2; // 2-d array of sub-components
  double* lambda;           // Probability of first state in logistic
  // double** sigma
  double** rho;             // Coefficients for linear terms in logistic
  // long is_singular;       // Flag whether the covariance matrix is singular
  long num_edges;           // The number of edges in the Chow-Liu forest
  double** ind_prob;        // Individual probabilities for each dimension
  long** edge;              // The list of edges (if any)
  double*** edge_prob;      // Corresponding edge probabilities
  long* edge_rank;          // Rank of the edge
  double* edge_MI;          // Corresponding mututal information (strength)
  double mdl_beta;          // Exponent modifier for the MDL prior
  class Envelope** envelope;// Clusters of edges in the dependence tree
  long num_envelopes;       // Number of clusters of edges
  long* node_used;          // The array indicating whether a node has been already used
  // double** sigma
  // double** inv_sigma
  double* first_mu;         // Mean for the first observation in the sequence
  double** first_sigma;     // Covariance for the first observation in the sequence
  double** inv_first_sigma; // The inverse of the covariance of first observations
  double** chol_first_sigma;// Upper-triangular Cholesky decomposition 
  double first_det;         // Determinant of the covariance matrix for the first state
  double** W;               // Transformation matrix
  // double* mu
  // double** sigma
  // double** inv_sigma
  // double det
  double** cond_state_prob; // Array of conditional probabilities for conditional (chain) Bernoulli
  double* mix_prob;         // Mixing probabilities for delta- distributions
  double* exp_param;        // Parameter used with delta-exponential distribution
  double* gamma_param1;     // Parameters used for delta-gamma distribution
  double* gamma_param2;
  /* Bivariate MaxEnt */
  // double** sigma -- correlation parameters
  //   with the diagonal for the univariate constraints
  // double det -- normalization constant
  long is_sim_initialized;  // Indicator whether the burn-off is already performed
  long* last_sim;           // Last simulated entry
  /* Tree structured Gaussian */
  // double** mu
  // double** sigma
  // double** inv_sigma
  // double det
  // double** chol_sigma
  // long num_edges
  // long** edge
  /* Dirac delta */
  double delta_value;
  /* Exponential (geometric) */
  double exp_param1;
  /* Gamma */
  double gamma1;
  double gamma2;
  /* Log-normal */
  double log_normal1;
  double log_normal2;

  // Univariate conditional exponential
  // double* lambda -- scaling parameters for the regression in the exponential
  // double* state_prob -- first sequence entry probabilities

  // Bayesian network decomposition bivariate MaxEnt
  // double** sigma -- parameters for the exponential model
  long** feature_index;     // Index for features for Bayesian networks MaxEnt
  long** feature_value;     // Values for feature functions for multinomial MaxEnt
  long* num_features;       // Number of features for each of the factors
  long* sim_order;          // Order in which the nodes need to be simulated

  /* Update variables */
  /* Conditional multinomial */
  double**** joint_prob;    // Joint probability of hidden states given evidence (for univariate)
  double**** uni_prob;      // Posterior univariate probabilities for all instances of hidden variables
  double*** fwd_update;     // Scaled results of the forward pass
  double*** bkd_update;     // Scaled results of the backward pass
  double** upd_scale;       // Inverse scaling factor
  double**** p_tr;          // P(S_nt=i|S_n,t-1=j,X_nt) -- transition probability (for univariate)
  double**** log_un_prob;   // log of unnormalized P(S_nt=i|S_n,t-1=j,X_nt) (for univariate)

  Distribution( long d_type, long d_num_states, long d_dim );
  ~Distribution();
  Distribution* copy();
  void Initialize( class Data* data );
  double prob( class DataPoint*, class DataPoint* );
  void probLogistic( class DataPoint*, long, double* );
  double log_prior();
  double log_un_probLogistic( class DataPoint*, long, long );
  void UpdateEmissionParameters( class Data*, double**, double );
  void UpdateTransitionParameters( class Data*, class Data*, long );
  void ReadParameters( FILE* );
  void ReadParameters2( FILE* );
  void UpdateIndex( long* );
  void special_logistic_NR( class Data*, class FlatData*, long, long, long, long );
  void general_logistic_NR( class Data*, double***, double****, double****, long, long, long );
  void linear_search_logistic_NR( Data*, double***, double****, double****, Distribution*, long );
  void linear_search_logistic_halving( Data*, double***, double****, double****, Distribution*, long, double, double );
  double logistic_derivative_along_vector( Data*, double***, double****, Distribution*, double );
  Distribution* logistic_gradient( Data*, double***, double****, double**** );
  void Normalize();
  void WriteToFile( FILE* );
  void WriteToFile2( FILE* );
  long num_params();
  void InformationGain( long, long**, double*, double, double*, long*, double*, double*, long );
  void Simulate( DataPoint*, DataPoint* );
  void SimulateMissingEntries( DataPoint*, DataPoint* );
  double weighted_ll( Data*, double** );
  long compute_envelopes( DataPoint*, long** );
  long compute_envelopes_full();

  void PostProcess();

  /* E-step auxiliary functions */
  void AllocateEMStructures( Data*, Data* );
  void AllocateForwardPassStructures( Data*, Data* );
  void CalculateTransitionValuesSequence( Sequence*, long, long );
  void CalculateForwardUpdatesSequence( Sequence*, long, long, double**** );
  void CalculateBackwardUpdatesSequence( Sequence*, long, double**** );
  void CalculateSummaries( Data*, double**** );
  void DeallocateEMStructures( Data*, Data* );
  void DeallocateForwardPassStructures( Data*, Data* );

  /* M-step updates */
  void UpdateEmissionMixture( Data*, double** );
  void UpdateEmissionCLMixture( Data*, double** );
  void UpdateEmissionBernoulli( Data*, double** );
  void UpdateEmissionConditionalBernoulli( Data*, double**, long );
  void UpdateEmissionChowLiu( Data*, double**, double );
  void UpdateEmissionConditionalChowLiu( Data*, double**, double );
  void UpdateEmissionFullBivariateMaxEnt( Data*, double**, double );
  void UpdateEmissionPUCMaxEnt( Data*, double**, double );
  void UpdateEmissionConditionalPUCMaxEnt( Data*, double**, double );
  void UpdateEmissionDeltaExponential( Data*, double**, double );
  void UpdateEmissionDeltaGamma( Data*, double**, double );
  void UpdateEmissionExponential( Data*, double** );
  void UpdateEmissionGamma( Data*, double** );
  void UpdateEmissionLognormal( Data*, double** );
  void UpdateEmissionGaussian( Data*, double**, double );
  void UpdateEmissionARGaussian( Data*, double**, double );
  void UpdateEmissionGaussianChowLiu( Data*, double**, double );
  void UpdateLogistic( Data*, double** );

  /* Expand/Project */
  Distribution* Project();
  void Expand( Distribution* );
};

class Envelope {
  /* Structure to keep track of chains of missing values */
public:
  long* node;
  long* edge;
  long num_nodes;
  long num_edges;

  Envelope( long );
  ~Envelope();
};

class DataPoint {
  /* Data point */
public:
  long ddim;       // Dimensionality of the discrete-valued variables
  long rdim;       // Dimensionality of the real-valued variables
  long* ddata;     // Discrete-valued data vector
  double* rdata;   // Real-valued data vector
  /* For simulation */
  bool* ddata_ind; // Indicator of whether the datum should be missing
  bool* rdata_ind; // Indicator of whether the datum should be missing

  DataPoint( long, long );
  ~DataPoint();
  void ReadData( FILE* );
  void WriteToFile( FILE* );
  long total_entries();
  void MissingEntryIndicators();
};

class Sequence {
  /* Class for a sequence of observations */
public:
  long seq_length;
  DataPoint** entry;
  long num_missing_discrete_entries;
  long** missing_discrete;

  Sequence( long );
  ~Sequence();
  void ReadData( FILE*, DataPoint* );
  void Simulate( Distribution**, long*, DataPoint* );
  void WriteToFile( FILE* );
  void UpdateMeanCounts( long*, long* );
  long total_entries();
  double weighted_total_entries( double* );
  void CatalogueMissingEntries();
  void MissingEntryIndicators();

private:
  long num_total_entries; // Number of initialized dimensions
};

class Data {
  /* Class for a data */
public:
  Sequence** sequence; /* Array of sequences */
  long num_seqs;
  char nametag[10];

  Data( long n_seqs=1, const char* tag="none\0", long is_p=0 );
  ~Data();
  void ReadData( FILE*, long, long, long, DataPoint* );
  void Simulate( Distribution**, long**, long, DataPoint* );
  void SimulateSL( Distribution*, Data*, DataPoint* );
  void WriteToFile( FILE* );
  long num_points();
  long total_entries();
  double weighted_total_entries( double** );
  double* mean();
  double*** covariance();
  double** persistence();
  long** spell( long );
  double** mutual_information();
  double** log_odds_ratio();
  void MissingEntryIndicators();
private:
  long is_primary;              // Indicator whether the set is primary or not
                                // Primary set points to sequences from other sets
                                // and need not deallocate its sequences
};

class StateSequence {
public:
  long* state[MAXBESTSEQ];         // Sequences collected
  double ll[MAXBESTSEQ];           // Log-likelihood of the sequence
  long seq_length;                 // Length of the sequences
  long num_best;                   // Number of best sequences collected
  long default_num_best;           // Number of best sequences requested

  StateSequence( long sl=1, long def=1 );
  ~StateSequence();
  void AddSequence( long*, double );
  void WriteToFile( FILE* );
};

class StateData {
  /* Class for state sequences */
public:
  StateSequence** sequence;
  long num_seqs;

  StateData( long );
  ~StateData();
  void WriteToFile( FILE* );
};


/* Functions */
long read_long( FILE* );
double read_double( FILE* );
long is_number_char( char );
char* read_word( FILE* );
long is_word_char( char );
void eat_blanks( FILE* );
void eat_line( FILE* );
long is_missing( long );
long is_missing( double );
long missing_value( long );
double missing_value( double );

Distribution* ReadDistribution( FILE* );

double find_det( double**, long );
void find_inv( double**, long, double** );
void cholesky( double**, long, double** );

long generateBernoulli( double*, long );
double generateUnitNormal( void );
void SimulateNormal( double*, double**, double*, long );
double generateExponential( double );
double generateGamma( double, double );
double generateLognormal( double, double );

double xlogx( double );
double gammaln( double );
double digamma( double );
double trigamma( double );
double gamma_dist( double, double, double );
double exp_dist( double, double );
double lognormal_dist( double, double, double );

void normalize_logistic( double*, double*, long, long );

double special_newton_raphson( double, double**, double**, double**, Data* );

double learn_univariate_conditional_maxent( long, long**, double*, double, long, double*, long, double*, long* );


void RunLearnParameters( Data*, Data*, HMM*, HMM**, Parameters*, FILE* );
void RunLogLikelihoodData( Data*, Data*, HMM*, FILE* );
void RunSimulateData( Data*, HMM*, Parameters*, FILE*, FILE*, long );
void RunHoleData( Data*, Data*, HMM*, FILE*, long );
void RunPredictionData( Data*, Data*, HMM*, long, FILE* );
void RunEvaluate( Data*, long, FILE* );
void TestDebug( Data*, Data*, HMM*, FILE* );

#endif
